#!/usr/bin/env python
# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logging.basicConfig(level=logging.DEBUG)
logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)

from spyne.application import Application
from spyne.decorator import rpc
from spyne.error import ResourceNotFoundError
from spyne.protocol.soap import Soap11
from spyne.model.primitive import Mandatory
from spyne.model.primitive import Unicode
from spyne.model.primitive import UnsignedInteger32
from spyne.model.complex import Array
from spyne.model.complex import Iterable
from spyne.model.complex import ComplexModel
from spyne.server.wsgi import WsgiApplication
from spyne.service import Service

_user_database = {}
_user_id_seq = 0


class Permission(ComplexModel):
    __namespace__ = 'spyne.examples.user_manager'

    id = UnsignedInteger32
    application = Unicode(values=('usermgr', 'accountmgr'))
    operation = Unicode(values=('read', 'modify', 'delete'))


class User(ComplexModel):
    __namespace__ = 'spyne.examples.user_manager'

    id = UnsignedInteger32
    user_name = Unicode(32, min_len=4, pattern='[a-z0-9.]+')
    full_name = Unicode(64, pattern='\w+( \w+)+')
    email = Unicode(pattern=r'[a-z0-9._%+-]+@[a-z0-9.-]+\.[A-Z]{2,4}')
    permissions = Array(Permission)


class UserManagerService(Service):
    @rpc(User.customize(min_occurs=1, nullable=False),
                                                     _returns=UnsignedInteger32)
    def put_user(ctx, user):
        if user.id is None:
            user.id = ctx.udc.get_next_user_id()
        ctx.udc.users[user.id] = user

        return user.id

    @rpc(Mandatory.UnsignedInteger32, _returns=User)
    def get_user(ctx, user_id):
        return ctx.udc.users[user_id]

    @rpc(Mandatory.UnsignedInteger32)
    def del_user(ctx, user_id):
        del ctx.udc.users[user_id]

    @rpc(_returns=Iterable(User))
    def get_all_users(ctx):
        return ctx.udc.users.itervalues()


class UserDefinedContext(object):
    def __init__(self):
        self.users = _user_database

    @staticmethod
    def get_next_user_id():
        global _user_id_seq

        _user_id_seq += 1

        return _user_id_seq

class MyApplication(Application):
    def call_wrapper(self, ctx):
        try:
            return super(MyApplication, self).call_wrapper(ctx)

        except KeyError:
            raise ResourceNotFoundError(ctx.in_object)


def _on_method_call(ctx):
    ctx.udc = UserDefinedContext()


application = MyApplication([UserManagerService],
            'spyne.examples.user_manager',
            in_protocol=Soap11(validator='lxml'),
            out_protocol=Soap11()
        )

application.event_manager.add_listener('method_call', _on_method_call)

def main():
    from wsgiref.simple_server import make_server

    server = make_server('127.0.0.1', 8000, WsgiApplication(application))

    logging.info("listening to http://127.0.0.1:8000")
    logging.info("wsdl is at: http://localhost:8000/?wsdl")

    return server.serve_forever()


if __name__=='__main__':
    import sys; sys.exit(main())




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

import getpass
import inspect
import traceback
import smtplib
import mimetypes

from socket import gethostname
from subprocess import Popen, PIPE

from email.utils import COMMASPACE, formatdate
from email import message_from_string
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.mime.message import MIMEMessage
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.nonmultipart import MIMENonMultipart
from email.encoders import encode_base64

from spyne.util import six


def email_exception(exception_address, message="", bcc=None):
    # http://stackoverflow.com/questions/1095601/find-module-name-of-the-originating-exception-in-python
    frm = inspect.trace()[-1]
    mod = inspect.getmodule(frm[0])
    module_name = mod.__name__ if mod else frm[1]

    sender = 'robot@spyne.io'
    recipients = [exception_address]
    if bcc is not None:
        recipients.extend(bcc)

    error_str = ("%s\n\n%s" % (message, traceback.format_exc()))
    msg = MIMEText(error_str.encode('utf8'), 'plain', 'utf8')
    msg['To'] = exception_address
    msg['From'] = 'Spyne <robot@spyne.io>'
    msg['Date'] = formatdate()
    msg['Subject'] = "(%s@%s) %s" % (getpass.getuser(), gethostname(), module_name)

    try:
        smtp_object = smtplib.SMTP('localhost')
        smtp_object.sendmail(sender, recipients, msg.as_string())
        logger.error("Error email sent")

    except Exception as e:
        logger.error("Error: unable to send email")
        logger.exception(e)


def email_text_smtp(addresses, sender=None, subject='', message="",
                                                     host='localhost', port=25):
    if sender is None:
        sender = 'Spyne <robot@spyne.io>'

    exc = traceback.format_exc()
    if exc is not None:
        message = (u"%s\n\n%s" % (message, exc))
    msg = MIMEText(message.encode('utf8'), 'plain', 'utf8')
    msg['To'] = COMMASPACE.join(addresses)
    msg['From'] = sender
    msg['Date'] = formatdate()
    msg['Subject'] = subject

    smtp_object = smtplib.SMTP(host, port)
    if six.PY2:
        smtp_object.sendmail(sender, addresses, msg.as_string())
    else:
        smtp_object.sendmail(sender, addresses, msg.as_bytes())
    logger.info("Text email sent to: %r.", addresses)


def email_text(addresses, sender=None, subject="", message="", bcc=None,
                                                                      att=None):
    if att is None:
        att = {}

    if sender is None:
        sender = 'Spyne <robot@spyne.io>'

    exc = traceback.format_exc()
    if exc is not None and exc != 'None\n' and exc != 'NoneType: None\n':
        message = (u"%s\n\n%s" % (message, exc))
    msg = MIMEText(message.encode('utf8'), 'plain', 'utf8')
    if len(att) > 0:
        newmsg = MIMEMultipart()
        newmsg.attach(msg)
        for k, v in att.items():
            mime_type, encoding = mimetypes.guess_type(k)
            if mime_type == "message/rfc822":
                part = MIMEMessage(message_from_string(v))
            elif mime_type.startswith("image/"):
                part = MIMEImage(v, mime_type.rsplit('/', 1)[-1])
            elif mime_type is not None:
                mime_type_main, mime_type_sub = mime_type.split('/', 1)
                part = MIMENonMultipart(mime_type_main, mime_type_sub)
                part.set_payload(v)
                encode_base64(part)
            else:
                part = MIMEApplication(v)

            newmsg.attach(part)
            part.add_header('Content-Disposition', 'attachment', filename=k)

        msg = newmsg

    msg['To'] = COMMASPACE.join(addresses)
    msg['From'] = sender
    msg['Date'] = formatdate()
    msg['Subject'] = subject

    cmd = ["/usr/sbin/sendmail", "-oi", '--']
    cmd.extend(addresses)
    if bcc is not None:
        cmd.extend(bcc)

    p = Popen(cmd, stdin=PIPE)
    if six.PY2:
        p.communicate(msg.as_string())
    else:
        p.communicate(msg.as_bytes())

    logger.info("Text email sent to: %r.", addresses)




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

import logging
logger = logging.getLogger('.'.join(__name__.split(".")[:-1]))

import os
import shutil
import tempfile

import spyne.const.xml as ns

from lxml import etree
from itertools import chain

from spyne.util.cdict import cdict
from spyne.util.odict import odict
from spyne.util.toposort import toposort2

from spyne.model import SimpleModel, ByteArray, ComplexModelBase, Fault, \
    Decimal, DateTime, Date, Time, Unicode
from spyne.model.enum import EnumBase
from spyne.interface import InterfaceDocumentBase

from spyne.interface.xml_schema.model import byte_array_add
from spyne.interface.xml_schema.model import simple_add
from spyne.interface.xml_schema.model import complex_add
from spyne.interface.xml_schema.model import fault_add
from spyne.interface.xml_schema.model import enum_add

from spyne.interface.xml_schema.model import simple_get_restriction_tag
from spyne.interface.xml_schema.model import unicode_get_restriction_tag
from spyne.interface.xml_schema.model import Tget_range_restriction_tag


_add_handlers = cdict({
    object: lambda interface, cls, tags: None,
    ByteArray: byte_array_add,
    SimpleModel: simple_add,
    ComplexModelBase: complex_add,
    Fault: fault_add,
    EnumBase: enum_add,
})

_get_restriction_tag_handlers = cdict({
    object: lambda self, cls: None,
    SimpleModel: simple_get_restriction_tag,
    Unicode: unicode_get_restriction_tag,
    Decimal: Tget_range_restriction_tag(Decimal),
    DateTime: Tget_range_restriction_tag(DateTime),
    Time: Tget_range_restriction_tag(Time),
    Date: Tget_range_restriction_tag(Date),
})

_ns_xsd = ns.NS_XSD
_ns_wsa = ns.NS_WSA
_ns_wsdl = ns.NS_WSDL11
_ns_soap = ns.NS_WSDL11_SOAP
_pref_wsa = ns.PREFMAP[_ns_wsa]


class SchemaInfo(object):
    def __init__(self):
        self.elements = odict()
        self.types = odict()


class XmlSchema(InterfaceDocumentBase):
    """The implementation of a subset of the Xml Schema 1.0 object definition
    document standard.

    The standard is available in three parts as follows:
    http://www.w3.org/TR/xmlschema-0/
    http://www.w3.org/TR/xmlschema-1/
    http://www.w3.org/TR/xmlschema-2/

    :param interface: A :class:`spyne.interface.InterfaceBase` instance.

    Supported events:
        * document_built:
            Called right after the document is built. The handler gets the
            ``XmlSchema`` instance as the only argument.

        * xml_document_built:
            Called right after the document is built. The handler gets the
            ``XmlSchema`` instance as the only argument. Only called from this
            class.
    """

    def __init__(self, interface, element_form_default='qualified'):
        super(XmlSchema, self).__init__(interface)

        self.element_form_default = element_form_default
        assert element_form_default in ('qualified', 'unqualified')

        self.schema_dict = {}
        self.validation_schema = None

        pref = self.interface.prefmap[self.interface.app.tns]
        self.namespaces = odict({pref: SchemaInfo()})

        self.complex_types = set()

    def add(self, cls, tags):
        if not (cls in tags):
            tags.add(cls)

            handler = _add_handlers[cls]
            handler(self, cls, tags)

    def get_restriction_tag(self, cls):
        handler = _get_restriction_tag_handlers[cls]
        return handler(self, cls)

    def build_schema_nodes(self, with_schema_location=False):
        self.schema_dict = {}

        tags = set()
        for cls in chain.from_iterable(toposort2(self.interface.deps)):
            self.add(cls, tags)

        for pref in self.namespaces:
            schema = self.get_schema_node(pref)

            # append import tags
            for namespace in self.interface.imports[self.interface.nsmap[pref]]:
                import_ = etree.SubElement(schema, ns.XSD('import'))

                import_.set("namespace", namespace)
                import_pref = self.interface.get_namespace_prefix(namespace)
                if with_schema_location and \
                                        self.namespaces.get(import_pref, False):
                    import_.set('schemaLocation', "%s.xsd" % import_pref)

                sl = ns.schema_location.get(namespace, None)
                if not (sl is None):
                    import_.set('schemaLocation', sl)

            # append simpleType and complexType tags
            for node in self.namespaces[pref].types.values():
                schema.append(node)

            # append element tags
            for node in self.namespaces[pref].elements.values():
                schema.append(node)

        self.add_missing_elements_for_methods()

        self.event_manager.fire_event('document_built', self)
        self.event_manager.fire_event('xml_document_built', self)

    def add_missing_elements_for_methods(self):
        def missing_methods():
            for service in self.interface.services:
                for method in service.public_methods.values():
                    if method.aux is None:
                        yield method

        pref_tns = self.interface.prefmap[self.interface.tns]

        elements = self.get_schema_info(pref_tns).elements
        schema_root = self.schema_dict[pref_tns]
        for method in missing_methods():
            name = method.in_message.Attributes.sub_name
            if name is None:
                name = method.in_message.get_type_name()

            if not name in elements:
                element = etree.Element(ns.XSD('element'))
                element.set('name', name)
                element.set('type', method.in_message.get_type_name_ns(
                                                                self.interface))
                elements[name] = element
                schema_root.append(element)

            if method.out_message is not None:
                name = method.out_message.Attributes.sub_name
                if name is None:
                    name = method.out_message.get_type_name()
                if not name in elements:
                    element = etree.Element(ns.XSD('element'))
                    element.set('name', name)
                    element.set('type', method.out_message \
                                              .get_type_name_ns(self.interface))
                    elements[name] = element
                    schema_root.append(element)

    def build_validation_schema(self):
        """Build application schema specifically for xml validation purposes."""

        self.build_schema_nodes(with_schema_location=True)

        pref_tns = self.interface.get_namespace_prefix(self.interface.tns)
        tmp_dir_name = tempfile.mkdtemp(prefix='spyne')
        logger.debug("generating schema for targetNamespace=%r, prefix: "
                  "%r in dir %r" % (self.interface.tns, pref_tns, tmp_dir_name))

        try:
            # serialize nodes to files
            for k, v in self.schema_dict.items():
                file_name = os.path.join(tmp_dir_name, "%s.xsd" % k)
                with open(file_name, 'wb') as f:
                    etree.ElementTree(v).write(f, pretty_print=True)

                logger.debug("writing %r for ns %s" %
                             (file_name, self.interface.nsmap[k]))

            with open(os.path.join(tmp_dir_name, "%s.xsd" % pref_tns), 'r') as f:
                try:
                    self.validation_schema = etree.XMLSchema(etree.parse(f))

                except Exception:
                    f.seek(0)
                    logger.error("This could be a Spyne error. Unless you're "
                                 "sure the reason for this error is outside "
                                 "Spyne, please open a new issue with a "
                                 "minimal test case that reproduces it.")
                    raise

            shutil.rmtree(tmp_dir_name)
            logger.debug("Schema built. Removed %r" % tmp_dir_name)

        except Exception as e:
            logger.exception(e)
            logger.error("The schema files are left at: %r" % tmp_dir_name)
            raise

    def get_schema_node(self, pref):
        """Return schema node for the given namespace prefix."""

        if not (pref in self.schema_dict):
            schema = etree.Element(ns.XSD('schema'),
                                                     nsmap=self.interface.nsmap)

            schema.set("targetNamespace", self.interface.nsmap[pref])
            schema.set("elementFormDefault", self.element_form_default)

            self.schema_dict[pref] = schema

        else:
            schema = self.schema_dict[pref]

        return schema

    def get_interface_document(self):
        return self.schema_dict

    def build_interface_document(self):
        self.build_schema_nodes()

    def add_element(self, cls, node):
        pref = cls.get_element_name_ns(self.interface).split(":")[0]

        schema_info = self.get_schema_info(pref)
        name = cls.Attributes.sub_name or cls.get_type_name()
        schema_info.elements[name] = node

    def add_simple_type(self, cls, node):
        tn = cls.get_type_name()
        pref = cls.get_namespace_prefix(self.interface)

        schema_info = self.get_schema_info(pref)
        schema_info.types[tn] = node

    def add_complex_type(self, cls, node):
        tn = cls.get_type_name()
        pref = cls.get_namespace_prefix(self.interface)

        schema_info = self.get_schema_info(pref)
        schema_info.types[tn] = node

    def get_schema_info(self, prefix):
        """Returns the SchemaInfo object for the corresponding namespace. It
        creates it if it doesn't exist.

        The SchemaInfo object holds the simple and complex type definitions
        for a given namespace."""

        if prefix in self.namespaces:
            schema = self.namespaces[prefix]
        else:
            schema = self.namespaces[prefix] = SchemaInfo()

        return schema



# encoding: utf8
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#


from __future__ import print_function

import logging
logger_c = logging.getLogger("%s.cloth" % __name__)
logger_s = logging.getLogger("%s.serializer" % __name__)

from lxml import html, etree
from copy import deepcopy
from inspect import isgenerator

from spyne.util import Break, coroutine
from spyne.util.oset import oset
from spyne.util.six import string_types
from spyne.util.color import R, B
from spyne.model import Array, AnyXml, AnyHtml, ModelBase, ComplexModelBase, \
    PushBase, XmlAttribute, AnyUri, XmlData, Any

from spyne.protocol import OutProtocolBase
from spyne.util.cdict import cdict

_revancestors = lambda elt: list(reversed(tuple(elt.iterancestors())))

_NODATA = type("_NODATA", (object,), {})


def _prevsibls(elt, strip_comments, since=None):
    return reversed(list(_prevsibls_since(elt, strip_comments, since)))


def _prevsibls_since(elt, strip_comments, since):
    if since is elt:
        return

    for prevsibl in elt.itersiblings(preceding=True):
        if prevsibl is since:
            break

        if strip_comments and isinstance(elt, etree.CommentBase):
            if elt.text.startswith('[if ') and elt.text.endswith('[endif]'):
                pass
            else:
                continue

        yield prevsibl


def _set_identifier_prefix(obj, prefix, mrpc_id='mrpc', id_attr='id',
                            data_tag='data', data_attr='data', attr_attr='attr',
                                        root_attr='root', tagbag_attr='tagbag'):
    obj.ID_PREFIX = prefix

    obj.MRPC_ID = '{}{}'.format(prefix, mrpc_id)
    obj.ID_ATTR_NAME = '{}{}'.format(prefix, id_attr)
    obj.DATA_TAG_NAME = '{}{}'.format(prefix, data_tag)
    obj.DATA_ATTR_NAME = '{}{}'.format(prefix, data_attr)
    obj.ATTR_ATTR_NAME = '{}{}'.format(prefix, attr_attr)
    obj.ROOT_ATTR_NAME = '{}{}'.format(prefix, root_attr)
    obj.TAGBAG_ATTR_NAME = '{}{}'.format(prefix, tagbag_attr)
    # FIXME: get rid of this. We don't want logic creep inside cloths
    obj.WRITE_CONTENTS_WHEN_NOT_NONE = '{}write-contents'.format(prefix)

    obj.SPYNE_ATTRS = {
        obj.ID_ATTR_NAME,
        obj.DATA_ATTR_NAME,
        obj.ATTR_ATTR_NAME,
        obj.ROOT_ATTR_NAME,
        obj.TAGBAG_ATTR_NAME,
        obj.WRITE_CONTENTS_WHEN_NOT_NONE,
    }


class ClothParserMixin(object):
    ID_PREFIX = 'spyne-'

    # these are here for documentation purposes. The are all reinitialized with
    # the call ta _set_identifier_prefix below the class definition
    ID_ATTR_NAME = 'spyne-id'
    DATA_TAG_NAME = 'spyne-data'
    DATA_ATTR_NAME = 'spyne-data'
    ATTR_ATTR_NAME = 'spyne-attr'
    ROOT_ATTR_NAME = 'spyne-root'
    TAGBAG_ATTR_NAME = 'spyne-tagbag'
    WRITE_CONTENTS_WHEN_NOT_NONE = 'spyne-write-contents'

    def set_identifier_prefix(self, what):
        _set_identifier_prefix(self, what)
        return self

    @classmethod
    def from_xml_cloth(cls, cloth, strip_comments=True):
        retval = cls()
        retval._init_cloth(cloth, cloth_parser=etree.XMLParser(),
                                                  strip_comments=strip_comments)
        return retval

    @classmethod
    def from_html_cloth(cls, cloth, strip_comments=True):
        retval = cls()
        retval._init_cloth(cloth, cloth_parser=html.HTMLParser(),
                                                  strip_comments=strip_comments)
        return retval

    @staticmethod
    def _strip_comments(root):
        for elt in root.iter():
            if isinstance(elt, etree.CommentBase):
                if elt.getparent() is not None:
                    if elt.text.startswith('[if ') \
                                               and elt.text.endswith('[endif]'):
                        pass
                    else:
                        elt.getparent().remove(elt)

    def _parse_file(self, file_name, cloth_parser):
        cloth = etree.parse(file_name, parser=cloth_parser)
        return cloth.getroot()

    def _init_cloth(self, cloth, cloth_parser, strip_comments):
        """Called from XmlCloth.__init__ in order to not break the dunder init
        signature consistency"""

        self._cloth = None
        self._root_cloth = None
        self.strip_comments = strip_comments

        self._mrpc_cloth = self._root_cloth = None

        if cloth is None:
            return

        if isinstance(cloth, string_types):
            cloth = self._parse_file(cloth, cloth_parser)

        if strip_comments:
            self._strip_comments(cloth)

        q = "//*[@%s]" % self.ROOT_ATTR_NAME
        elts = cloth.xpath(q)
        if len(elts) > 0:
            logger_c.debug("Using %r as root cloth.", cloth)
            self._root_cloth = elts[0]
        else:
            logger_c.debug("Using %r as plain cloth.", cloth)
            self._cloth = cloth

        self._mrpc_cloth = self._pop_elt(cloth, 'mrpc_entry')

    def _pop_elt(self, elt, what):
        query = '//*[@%s="%s"]' % (self.ID_ATTR_NAME, what)
        retval = elt.xpath(query)
        if len(retval) > 1:
            raise ValueError("more than one element found for query %r" % query)

        elif len(retval) == 1:
            retval = retval[0]
            next(retval.iterancestors()).remove(retval)
            return retval


_set_identifier_prefix(ClothParserMixin, ClothParserMixin.ID_PREFIX)


class ToClothMixin(OutProtocolBase, ClothParserMixin):
    def __init__(self, app=None, mime_type=None, ignore_uncap=False,
                                        ignore_wrappers=False, polymorphic=True):
        super(ToClothMixin, self).__init__(app=app, mime_type=mime_type,
                     ignore_uncap=ignore_uncap, ignore_wrappers=ignore_wrappers)

        self.polymorphic = polymorphic
        self.rendering_handlers = cdict({
            ModelBase: self.model_base_to_cloth,
            AnyXml: self.xml_to_cloth,
            Any: self.any_to_cloth,
            AnyHtml: self.html_to_cloth,
            AnyUri: self.any_uri_to_cloth,
            ComplexModelBase: self.complex_to_cloth,
        })

    def _get_elts(self, elt, tag_id=None):
        if tag_id is None:
            return elt.xpath('.//*[@*[starts-with(name(), "%s")]]' %
                                                                 self.ID_PREFIX)
        return elt.xpath('.//*[@*[starts-with(name(), "%s")]="%s"]' % (
                                                        self.ID_PREFIX, tag_id))

    def _get_outmost_elts(self, tmpl, tag_id=None):
        ids = set()

        # we assume xpath() returns elements in top to bottom (or outside to
        # inside) order.
        for elt in self._get_elts(tmpl, tag_id):
            if elt is tmpl:  # FIXME: kill this
                logger_c.debug("Don't send myself")
                continue  # don't send myself

            if len(set((id(e) for e in elt.iterancestors())) & ids) > 0:
                logger_c.debug("Don't send grandchildren")
                continue  # don't send grandchildren

            if id(elt) in ids:  # FIXME: this check should be safe to remove
                logger_c.debug("Don't send what's already been sent")
                continue  # don't send what's already been sent

            if self.ID_ATTR_NAME in elt.attrib:
                # Prevent primitive attrs like spyne-attr from interfering
                # with elt descent
                ids.add(id(elt))

            yield elt

    def _get_clean_elt(self, elt, what):
        query = '//*[@%s="%s"]' % (self.ID_ATTR_NAME, what)
        retval = elt.xpath(query)
        if len(retval) > 1:
            raise ValueError("more than one element found for query %r" % query)

        elif len(retval) == 1:
            retval = retval[0]
            del retval.attrib[self.ID_ATTR_NAME]
            return retval

    def _get_elts_by_id(self, elt, what):
        retval = elt.xpath('//*[@id="%s"]' % what)
        logger_c.debug("id=%r got %r", what, retval)
        return retval

    def _is_tagbag(self, elt):
        return self.TAGBAG_ATTR_NAME in elt.attrib

    @staticmethod
    def _methods(ctx, cls, inst):
        while cls.Attributes._wrapper and len(cls._type_info) > 0:
            cls, = cls._type_info.values()

        if cls.Attributes.methods is not None:
            for k, v in cls.Attributes.methods.items():
                is_shown = True
                if v.when is not None:
                    is_shown = v.when(inst, ctx)

                if is_shown:
                    yield k, v

    def _actions_to_cloth(self, ctx, cls, inst, template):
        if self._mrpc_cloth is None:
            logger_c.warning("missing 'mrpc_template'")
            return

        for elt in self._get_elts(template, self.MRPC_ID):
            for k, v in self._methods(ctx, cls, inst):
                href = v.in_message.get_type_name()
                text = v.translate(ctx.locale, v.in_message.get_type_name())

                mrpc_template = deepcopy(self._mrpc_cloth)
                anchor = self._get_clean_elt(mrpc_template, 'mrpc_link')
                anchor.attrib['href'] = href

                text_elt = self._get_clean_elt(mrpc_template, 'mrpc_text')
                if text_elt is not None:
                    text_elt.text = text
                else:
                    anchor.text = text

                elt.append(mrpc_template)
                                           # mutable default ok because readonly
    def _enter_cloth(self, ctx, cloth, parent, attrib={}, skip=False,
                                                  method=None, skip_dupe=False):
        """Enters the given tag in the document by using the shortest path from
        current tag.

        1. Moves up the tree by writing all tags so that the set of ancestors
           of the current tag are a subset of the ancestors of the parent tag
        2. Writes all tags until hitting a direct ancestor, enters it, and
           keeps writing previous siblings of ancestor tags and entering
           ancestor tags until hitting the target tag.
        3. Enters the target tag and returns

        There is no _exit_cloth because exiting from tags is done
        automatically with subsequent calls to _enter_cloth and finally to
        _close_cloth.

        :param ctx: A MethodContext instance
        :param cloth: The target cloth -- an ``lxml.etree._Element`` instance.
        :param parent: The target stream -- typically an
            ``lxml.etree._IncrementalFileWriter`` instance.
        :param attrib: A dict of additional attributes for the target cloth.
        :param skip: When True, the target tag is actually not entered.
            Typically used for XmlData and friends.
        :param method: One of ``(None, 'html', 'xml')``. When not ``None``,
            overrides the output method of lxml.
        :param skip_dupe: When ``False`` (the default) if this function is
            called repeatedly for the same tag, the tag is exited and reentered.
            This typically happens for types with ``max_occurs`` > 1
            (eg. arrays).
        """

        logger_c.debug("entering %s %r nsmap=%r attrib=%r skip=%s method=%s",
                     cloth.tag, cloth.attrib, cloth.nsmap, attrib, skip, method)

        if not ctx.outprot_ctx.doctype_written:
            self.write_doctype(ctx, parent, cloth)

        tags = ctx.protocol.tags
        rootstack = ctx.protocol.rootstack
        assert isinstance(rootstack, oset)

        eltstack = ctx.protocol.eltstack
        ctxstack = ctx.protocol.ctxstack

        cureltstack = eltstack[rootstack.back]
        curctxstack = ctxstack[rootstack.back]

        if skip_dupe and len(cureltstack) > 0 and cureltstack[-1] is cloth:
            return

        cloth_root = cloth.getroottree().getroot()
        if not cloth_root in rootstack:
            rootstack.add(cloth_root)
            cureltstack = eltstack[rootstack.back]
            curctxstack = ctxstack[rootstack.back]

            assert rootstack.back == cloth_root

        while rootstack.back != cloth_root:
            self._close_cloth(ctx, parent)

        last_elt = None
        if len(cureltstack) > 0:
            last_elt = cureltstack[-1]

        ancestors = _revancestors(cloth)

        # move up in tag stack until the ancestors of both
        # source and target tags match
        while ancestors[:len(cureltstack)] != cureltstack:
            elt = cureltstack.pop()
            elt_ctx = curctxstack.pop()

            last_elt = elt
            if elt_ctx is not None:
                self.event_manager.fire_event(("before_exit", elt), ctx, parent)
                elt_ctx.__exit__(None, None, None)
                logger_c.debug("\texit norm %s %s", elt.tag, elt.attrib)
                if elt.tail is not None:
                    parent.write(elt.tail)

            # unless we're at the same level as the relevant ancestor of the
            # target node
            if ancestors[:len(cureltstack)] != cureltstack:
                # write following siblings before closing parent node
                for sibl in elt.itersiblings(preceding=False):
                    logger_c.debug("\twrite exit sibl %s %r %d",
                                                sibl.tag, sibl.attrib, id(sibl))
                    parent.write(sibl)

        # write remaining ancestors of the target node.
        for anc in ancestors[len(cureltstack):]:
            # write previous siblings of ancestors (if any)
            prevsibls = _prevsibls(anc, self.strip_comments, since=last_elt)
            for elt in prevsibls:
                if id(elt) in tags:
                    logger_c.debug("\tskip  anc prevsibl %s %r",
                                                            elt.tag, elt.attrib)
                    continue

                logger_c.debug("\twrite anc prevsibl %s %r 0x%x",
                                                   elt.tag, elt.attrib, id(elt))
                parent.write(elt)

            # enter the ancestor node
            kwargs = {}
            if len(cureltstack) == 0:
                # if this is the first node ever, initialize namespaces as well
                kwargs['nsmap'] = anc.nsmap

            anc_ctx = parent.element(anc.tag, anc.attrib, **kwargs)
            anc_ctx.__enter__()
            logger_c.debug("\tenter norm %s %r 0x%x method: %r", anc.tag,
                                                    anc.attrib, id(anc), method)
            if anc.text is not None:
                parent.write(anc.text)

            rootstack.add(anc.getroottree().getroot())
            cureltstack = eltstack[rootstack.back]
            curctxstack = ctxstack[rootstack.back]
            cureltstack.append(anc)
            curctxstack.append(anc_ctx)

        # now that at the same level as the target node,
        # write its previous siblings
        prevsibls = _prevsibls(cloth, self.strip_comments, since=last_elt)
        for elt in prevsibls:
            if elt is last_elt:
                continue

            if id(elt) in tags:
                logger_c.debug("\tskip  cloth prevsibl %s %r",
                                                            elt.tag, elt.attrib)
                continue

            logger_c.debug("\twrite cloth prevsibl %s %r", elt.tag, elt.attrib)
            parent.write(elt)

        skip = skip or (cloth.tag == self.DATA_TAG_NAME)

        if skip:
            tags.add(id(cloth))
            if method is not None:
                curtag = parent.method(method)
                curtag.__enter__()
            else:
                curtag = None

        else:
            # finally, enter the target node.
            cloth_attrib = dict([(k, v) for k, v in cloth.attrib.items()
                                                  if not k in self.SPYNE_ATTRS])

            cloth_attrib.update(attrib)

            self.event_manager.fire_event(("before_entry", cloth), ctx,
                                                           parent, cloth_attrib)

            kwargs = {}
            if len(cureltstack) == 0:
                # if this is the first node ever, initialize namespaces as well
                kwargs['nsmap'] = cloth.nsmap
            if method is not None:
                kwargs['method'] = method
            curtag = parent.element(cloth.tag, cloth_attrib, **kwargs)
            curtag.__enter__()
            if cloth.text is not None:
                parent.write(cloth.text)

        rootstack.add(cloth.getroottree().getroot())
        cureltstack = eltstack[rootstack.back]
        curctxstack = ctxstack[rootstack.back]

        cureltstack.append(cloth)
        curctxstack.append(curtag)

        logger_c.debug("")

    def _close_cloth(self, ctx, parent):
        rootstack = ctx.protocol.rootstack
        close_until = rootstack.back
        cureltstack = ctx.protocol.eltstack[close_until]
        curctxstack = ctx.protocol.ctxstack[close_until]

        for elt, elt_ctx in reversed(tuple(zip(cureltstack, curctxstack))):
            if elt_ctx is not None:
                self.event_manager.fire_event(("before_exit", elt), ctx, parent)
                elt_ctx.__exit__(None, None, None)
                logger_c.debug("exit %s close", elt.tag)
                if elt.tail is not None:
                    parent.write(elt.tail)

            for sibl in elt.itersiblings(preceding=False):
                logger_c.debug("write %s nextsibl", sibl.tag)
                parent.write(sibl)
                if sibl.tail is not None:
                    parent.write(sibl.tail)

            if elt is close_until:
                logger_c.debug("closed until %r, breaking out", close_until)
                break

        del ctx.protocol.eltstack[close_until]
        del ctx.protocol.ctxstack[close_until]

        if len(rootstack) > 0:
            rootstack.pop()

    @coroutine
    def to_parent_cloth(self, ctx, cls, inst, cloth, parent, name,
                                                      from_arr=False, **kwargs):
        cls_cloth = self.get_class_cloth(cls)
        if cls_cloth is not None:
            logger_c.debug("%r to object cloth", cls)
            cloth = cls_cloth
            ctx.protocol[self].rootstack.add(cloth)

        ret = self.to_cloth(ctx, cls, inst, cloth, parent, '')
        if isgenerator(ret):
            try:
                while True:
                    sv2 = (yield)
                    ret.send(sv2)
            except Break as e:
                try:
                    ret.throw(e)
                except (Break, StopIteration, GeneratorExit):
                    pass

    @coroutine
    def to_root_cloth(self, ctx, cls, inst, cloth, parent, name):
        if len(ctx.protocol.eltstack) > 0:
            ctx.protocol[self].rootstack.add(cloth)

        cls_attrs = self.get_cls_attrs(cls)
        self._enter_cloth(ctx, cloth, parent, method=cls_attrs.method)

        ret = self.start_to_parent(ctx, cls, inst, parent, name)
        if isgenerator(ret):
            try:
                while True:
                    sv2 = (yield)
                    ret.send(sv2)
            except Break as e:
                try:
                    ret.throw(e)
                except (Break, StopIteration, GeneratorExit):
                    pass

    # TODO: Maybe DRY this with to_parent?
    @coroutine
    def to_cloth(self, ctx, cls, inst, cloth, parent, name=None,
                        from_arr=False, as_attr=False, as_data=False, **kwargs):

        prot_name = self.__class__.__name__

        if issubclass(cls, XmlAttribute):
            cls = cls.type
            as_attr = True

        elif issubclass(cls, XmlData):
            cls = cls.type
            as_data = True

        pushed = False
        if cloth is None:
            logger_c.debug("No cloth fround, switching to to_parent...")
            ret = self.to_parent(ctx, cls, inst, parent, name, **kwargs)

        else:
            cls, _ = self.get_polymorphic_target(cls, inst)
            cls_attrs = self.get_cls_attrs(cls)

            inst = self._sanitize(cls_attrs, inst)

            # if instance is None, use the default factory to generate one
            _df = cls_attrs.default_factory
            if inst is None and callable(_df):
                inst = _df()

            # if instance is still None, use the default value
            if inst is None:
                inst = cls_attrs.default

            # if there's a subprotocol, switch to it
            subprot = cls_attrs.prot
            if subprot is not None and not (subprot is self):
                # we can't do this because subprotocols don't accept cloths.
                # so we need to enter the cloth, which make it too late to
                # set attributes.
                assert not as_attr, "No subprot supported for fields " \
                    "to be serialized as attributes, use type casting with "  \
                    "customized serializers in the current protocol instead."

                self._enter_cloth(ctx, cloth, parent,
                                          method=cls_attrs.method, skip=as_data)

                ret = subprot.subserialize(ctx, cls, inst, parent, name,
                                     as_attr=as_attr, as_data=as_data, **kwargs)

            # if there is no subprotocol, try rendering the value
            else:
                ret = None

                # try rendering the null value
                if inst is None:
                    if cls_attrs.min_occurs > 0:
                        attrs = {}
                        if as_attr:
                            # FIXME: test needed
                            attrs[name] = ''

                        self._enter_cloth(ctx, cloth, parent, attrib=attrs,
                                                        method=cls_attrs.method)
                        identifier = "%s.%s" % (prot_name, "null_to_cloth")
                        logger_s.debug("Writing '%s' using %s type: %s.", name,
                                                identifier, cls.get_type_name())
                        parent.write(cloth)

                    else:
                        logger_s.debug("Skipping '%s' type: %s because empty.",
                                                      name, cls.get_type_name())
                        self._enter_cloth(ctx, cloth, parent, skip=True,
                                                        method=cls_attrs.method)

                elif as_data:
                    # we only support XmlData of a primitive.,. is this a
                    # problem?
                    ret = self.to_unicode(cls, inst)
                    if ret is not None:
                        parent.write(ret)

                elif as_attr:
                    sub_name = cls_attrs.sub_name
                    if sub_name is None:
                        sub_name = name
                    attrs = {sub_name: self.to_unicode(cls, inst)}

                    self._enter_cloth(ctx, cloth, parent, attrib=attrs,
                                                        method=cls_attrs.method)

                else:
                    # push the instance at hand to instance stack. this makes it
                    # easier for protocols to make decisions based on parents of
                    # instances at hand.
                    pushed = True
                    logger_c.debug("%s %r pushed %r %r", R("#"), self, cls, inst)
                    ctx.outprot_ctx.inst_stack.append((cls, inst, from_arr))

                    # try rendering the array value
                    if not from_arr and cls.Attributes.max_occurs > 1:
                        ret = self.array_to_cloth(ctx, cls, inst, cloth, parent,
                                                     as_attr=as_attr, name=name)
                    else:
                        # try rendering anything else
                        handler = self.rendering_handlers[cls]

                        # disabled for performance reasons
                        # identifier = "%s.%s" % (prot_name, handler.__name__)
                        # from spyne.util.web import log_repr
                        # logger_s.debug("Writing %s using %s for %s. Inst: %r",
                        #              name, identifier, cls.get_type_name(),
                        #              log_repr(inst, cls, from_array=from_arr))

                        ret = handler(ctx, cls, inst, cloth, parent, name=name,
                                                                as_attr=as_attr)

        if isgenerator(ret):
            try:
                while True:
                    sv2 = (yield)
                    ret.send(sv2)
            except Break as e:
                try:
                    ret.throw(e)
                except (Break, StopIteration, GeneratorExit):
                    pass
                finally:
                    if pushed:
                        logger_c.debug("%s %r popped %r %r", B("#"),
                                                                self, cls, inst)
                        ctx.outprot_ctx.inst_stack.pop()

        else:
            if pushed:
                logger_c.debug("%s %r popped %r %r", B("#"), self, cls, inst)
                ctx.outprot_ctx.inst_stack.pop()

    def model_base_to_cloth(self, ctx, cls, inst, cloth, parent, name,
                                                                      **kwargs):
        cls_attrs = self.get_cls_attrs(cls)
        self._enter_cloth(ctx, cloth, parent, method=cls_attrs.method)

        # FIXME: Does it make sense to do this in other types?
        if self.WRITE_CONTENTS_WHEN_NOT_NONE in cloth.attrib:
            logger_c.debug("Writing contents for %r", cloth)
            for c in cloth:
                parent.write(c)

        else:
            parent.write(self.to_unicode(cls, inst))

    def xml_to_cloth(self, ctx, cls, inst, cloth, parent, name, **_):
        cls_attrs = self.get_cls_attrs(cls)
        self._enter_cloth(ctx, cloth, parent, method=cls_attrs.method)
        if isinstance(inst, string_types):
            inst = etree.fromstring(inst)
        parent.write(inst)

    def any_to_cloth(self, ctx, cls, inst, cloth, parent, name, **_):
        cls_attrs = self.get_cls_attrs(cls)
        self._enter_cloth(ctx, cloth, parent, method=cls_attrs.method)
        parent.write(inst)

    def html_to_cloth(self, ctx, cls, inst, cloth, parent, name, **_):
        cls_attrs = self.get_cls_attrs(cls)
        self._enter_cloth(ctx, cloth, parent, method=cls_attrs.method)
        if isinstance(inst, string_types):
            inst = html.fromstring(inst)
        parent.write(inst)

    def any_uri_to_cloth(self, ctx, cls, inst, cloth, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)
        self._enter_cloth(ctx, cloth, parent, method=cls_attrs.method)
        self.any_uri_to_parent(ctx, cls, inst, parent, name, **kwargs)

    @coroutine
    def complex_to_cloth(self, ctx, cls, inst, cloth, parent, name=None,
                                                       as_attr=False, **kwargs):
        fti = cls.get_flat_type_info(cls)
        cls_attrs = self.get_cls_attrs(cls)

        # It's actually an odict but that's irrelevant here.
        fti_check = dict(fti.items())
        elt_check = set()

        attrib = self._gen_attrib_dict(inst, fti)
        self._enter_cloth(ctx, cloth, parent, attrib=attrib,
                                                        method=cls_attrs.method)

        for elt in self._get_elts(cloth, self.MRPC_ID):
            self._actions_to_cloth(ctx, cls, inst, elt)

        if self._is_tagbag(cloth):
            logger_c.debug("%r(%r) IS a tagbag", cloth, cloth.attrib)
            elts = self._get_elts(cloth)
        else:
            logger_c.debug("%r(%r) is NOT a tagbag", cloth, cloth.attrib)
            elts = self._get_outmost_elts(cloth)

        # Check for xmldata after entering the cloth.
        as_data_field = cloth.attrib.get(self.DATA_ATTR_NAME, None)
        if as_data_field is not None:
            self._process_field(ctx, cls, inst, parent, cloth, fti,
                   as_data_field, as_attr, True, fti_check, elt_check, **kwargs)

        for elt in elts:
            for k_attr, as_attr, as_data in ((self.ID_ATTR_NAME, False, False),
                                            (self.ATTR_ATTR_NAME, True, False),
                                            (self.DATA_ATTR_NAME, False, True)):
                field_name = elt.attrib.get(k_attr, None)
                if field_name is None:
                    continue

                if elt.tag == self.DATA_TAG_NAME:
                    as_data = True

                ret = self._process_field(ctx, cls, inst, parent, elt, fti,
                     field_name, as_attr=as_attr, as_data=as_data,
                             fti_check=fti_check, elt_check=elt_check, **kwargs)

                if isgenerator(ret):
                    try:
                        while True:
                            sv2 = (yield)
                            ret.send(sv2)
                    except Break as e:
                        try:
                            ret.throw(e)
                        except StopIteration:
                            pass
                        finally:
                            # cf below
                            if not (as_attr or as_data):
                                break
                else:
                    # this is here so that attribute on complex model doesn't get
                    # mixed with in-line attr inside complex model. if an element
                    # has spyne-id, all other attrs are ignored and are processed
                    # by the object's serializer not its parent.
                    if not (as_attr or as_data):
                        break

        if len(fti_check) > 0:
            logger_s.debug("No element found for the following fields: %r",
                                                         list(fti_check.keys()))
        if len(elt_check) > 0:
            logger_s.debug("No field found for element the following "
                                              "elements: %r", list(elt_check))

    def _process_field(self, ctx, cls, inst, parent,
                   elt, fti, field_name, as_attr, as_data, fti_check, elt_check,
                                                                      **kwargs):
        field_type = fti.get(field_name, None)
        fti_check.pop(field_name, None)

        if field_type is None:
            logger_c.warning("elt id %r not in %r", field_name, cls)
            elt_check.add(field_name)
            self._enter_cloth(ctx, elt, parent, skip=True)
            return

        cls_attrs = self.get_cls_attrs(field_type)
        if cls_attrs.exc:
            logger_c.debug("Skipping elt id %r because "
                           "it was excluded", field_name)
            return

        sub_name = cls_attrs.sub_name
        if sub_name is None:
            sub_name = field_name

        if issubclass(cls, Array):
            # if cls is an array, inst should already be a sequence type
            # (eg list), so there's no point in doing a getattr -- we will
            # unwrap it and serialize it in the next round of to_cloth call.
            val = inst
        else:
            val = getattr(inst, field_name, None)

        if as_data:
            self._enter_cloth(ctx, elt, parent, skip=True, skip_dupe=True,
                                                        method=cls_attrs.method)

        return self.to_cloth(ctx, field_type, val, elt, parent,
                      name=sub_name, as_attr=as_attr, as_data=as_data, **kwargs)

    @coroutine
    def array_to_cloth(self, ctx, cls, inst, cloth, parent, name=None,
                                                                      **kwargs):
        if isinstance(inst, PushBase):
            while True:
                sv = (yield)
                ret = self.to_cloth(ctx, cls, sv, cloth, parent,
                                             name=name, from_arr=True, **kwargs)
                if isgenerator(ret):
                    try:
                        while True:
                            sv2 = (yield)
                            ret.send(sv2)
                    except Break as e:
                        try:
                            ret.throw(e)
                        except StopIteration:
                            pass

        else:
            sv = _NODATA

            for sv in inst:
                was_empty = False

                ret = self.to_cloth(ctx, cls, sv, cloth, parent,
                                             from_arr=True, name=name, **kwargs)
                if isgenerator(ret):
                    try:
                        while True:
                            sv2 = (yield)
                            ret.send(sv2)
                    except Break as e:
                        try:
                            ret.throw(e)
                        except StopIteration:
                            pass

            if sv is _NODATA:
                # FIXME: what if min_occurs >= 1?
                # fake entering the cloth to prevent it from being flushed as
                # parent or sibling of another node later.
                self._enter_cloth(ctx, cloth, parent, skip=True)




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""The module for memoization stuff.

When you have memory leaks in your daemon, the reason could very well be
reckless usage of the tools here.

These are NOT thread-safe. If you are relying on exactly-one-execution-per-key
behavior in a multithreaded environment, roll your own stuff.
"""


import logging
logger = logging.getLogger(__name__)

import threading


MEMOIZATION_STATS_LOG_INTERVAL = 60.0


def _log_all():
    logger.info("%d memoizers", len(memoize.registry))
    for memo in memoize.registry:
        logger.info("%r: %d entries.", memo.func, len(memo.memo))


def _log_func(func):
    for memo in memoize.registry:
        if memo.func is func.func.im_self.func:
            break
    else:
        logger.error("%r not found in memoization regisry", func)
        return

    logger.info("%r: %d entries.", memo.func, len(memo.memo))
    for k, v in memo.memo.items():
        logger.info("\t%r: %r", k, v)


def start_memoization_stats_logger(func=None):
    logger.info("Enabling @memoize statistics every %d second(s).",
                                                 MEMOIZATION_STATS_LOG_INTERVAL)

    if func is None:
        _log_all()
    else:
        _log_func(func)

    t = threading.Timer(MEMOIZATION_STATS_LOG_INTERVAL,
                                        start_memoization_stats_logger, (func,))

    t.daemon = True
    t.start()


class memoize(object):
    """A memoization decorator that keeps caching until reset."""

    registry = []

    def __init__(self, func):
        self.func = func
        self.memo = {}
        self.lock = threading.RLock()
        memoize.registry.append(self)

    def __call__(self, *args, **kwargs):
        key = self.get_key(args, kwargs)
        # we hope that gil makes this comparison is race-free
        if not key in self.memo:
            with self.lock:
                # make sure the situation hasn't changed after lock acq
                if not key in self.memo:
                    value = self.func(*args, **kwargs)
                    self.memo[key] = value
                    return value
        return self.memo.get(key)

    def get_key(self, args, kwargs):
        return tuple(args), tuple(kwargs.items())

    def reset(self):
        self.memo = {}


class memoize_first(object):
    """A memoization decorator that keeps the first call without condition, aka
    a singleton accessor."""

    registry = []

    def __init__(self, func):
        self.func = func
        self.lock = threading.RLock()
        memoize.registry.append(self)

    def __call__(self, *args, **kwargs):
        if not hasattr(self, 'memo'):
            value = self.func(*args, **kwargs)
            self.memo = value
            return value
        return self.memo

    def reset(self):
        del self.memo


def memoize_ignore(values):
    """A memoization decorator that does memoization unless the returned
    value is in the 'values' iterable. eg let `values = (2,)` and
    `add = lambda x, y: x + y`, the result of `add(1, 1)` (=2) is not memoized
    but the result of `add(5, 5)` (=10) is.
    """

    assert iter(values), \
                       "memoize_ignore requires an iterable of values to ignore"

    class _memoize_ignored(memoize):
        def __call__(self, *args, **kwargs):
            key = self.get_key(args, kwargs)
            # we hope that gil makes this comparison is race-free
            if not key in self.memo:
                with self.lock:
                    # make sure the situation hasn't changed after lock acq
                    if not key in self.memo:
                        value = self.func(*args, **kwargs)
                        if not value in values:
                            self.memo[key] = value

                        return value
            return self.memo.get(key)

    return _memoize_ignored


class memoize_ignore_none(memoize):
    """A memoization decorator that ignores `None` values. ie when the decorated
    function returns `None`, the value is returned but not memoized.
    """

    def __call__(self, *args, **kwargs):
        key = self.get_key(args, kwargs)
        # we hope that gil makes this comparison is race-free
        if not key in self.memo:
            with self.lock:
                # make sure the situation hasn't changed after lock acq
                if not key in self.memo:
                    value = self.func(*args, **kwargs)
                    if not (value is None):
                        self.memo[key] = value

                    return value
        return self.memo.get(key)


class memoize_id(memoize):
    """A memoization decorator that keeps caching until reset for unhashable
    types. It works on id()'s of objects instead."""

    def get_key(self, args, kwargs):
        return tuple([id(a) for a in args]), \
                                  tuple([(k, id(v)) for k, v in kwargs.items()])




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""
This module contains an ordered dictionary implementation.

We need this in Python 2.7 because collections.OrderedDict does not support
reordering by assignment to keys().

We need this in Python 3.x because keys() returns KeyView which which doesn't
support `__getitem__` -- i.e. getting nth variable from the ordered dict.
"""


class odict(dict):
    """Sort of an ordered dictionary implementation."""

    def __init__(self, data=()):
        if isinstance(data, self.__class__):
            self.__list = list(data.__list)
            super(odict, self).__init__(data)

        else:
            self.__list = []
            super(odict, self).__init__()
            self.update(data)

    def __getitem__(self, key):
        if isinstance(key, int):
            return super(odict, self).__getitem__(self.__list[key])
        else:
            return super(odict, self).__getitem__(key)

    def __setitem__(self, key, val):
        if isinstance(key, int):
            super(odict, self).__setitem__(self.__list[key], val)

        else:
            if not (key in self):
                self.__list.append(key)
            super(odict, self).__setitem__(key, val)

        assert len(self.__list) == super(odict, self).__len__(), (
            repr(self.__list), super(odict, self).__repr__())

    def __repr__(self):
        return "{%s}" % ','.join(["%r: %r" % (k, v) for k, v in self.items()])

    def __str__(self):
        return repr(self)

    def __len__(self):
        assert len(self.__list) == super(odict, self).__len__()
        return len(self.__list)

    def __iter__(self):
        return iter(self.__list)

    def __delitem__(self, key):
        if not isinstance(key, int):
            super(odict, self).__delitem__(key)
            key = self.__list.index(key) # ouch.
        else:
            super(odict, self).__delitem__(self.__list[key])
        del self.__list[key]

    def __add__(self, other):
        self.update(other)
        return self

    def items(self):
        retval = []
        for k in self.__list:
            retval.append( (k, super(odict, self).__getitem__(k)) )
        return retval

    def iteritems(self):
        for k in self.__list:
            yield k, super(odict, self).__getitem__(k)

    def keys(self):
        return self.__list

    def update(self, data, **kwargs):
        if isinstance(data, (dict, odict)):
            data = data.items()

        for k, v in data:
            self[k] = v

        for k, v in kwargs.items():
            self[k] = v

    def values(self):
        retval = []
        for l in self.__list:
            retval.append(super(odict, self).__getitem__(l))
        return retval

    def itervalues(self):
        for l in self.__list:
            yield self[l]

    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default

    def append(self, t):
        k, v = t
        self[k] = v

    def insert(self, index, item):
        k, v = item
        if k in self:
            del self.__list[self.__list.index(k)]
        self.__list.insert(index, k)
        super(odict, self).__setitem__(k, v)




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

from datetime import datetime
from weakref import WeakKeyDictionary

from spyne import ProtocolContext, EventManager
from spyne.const import DEFAULT_LOCALE
from spyne.model import Array
from spyne.error import ResourceNotFoundError
from spyne.util import DefaultAttrDict
from spyne.util.six import string_types


_MISSING = type("_MISSING", (object,), {})()


class ProtocolMixin(object):
    mime_type = 'application/octet-stream'

    SOFT_VALIDATION = type("Soft", (object,), {})
    REQUEST = type("Request", (object,), {})
    RESPONSE = type("Response", (object,), {})

    type = set()
    """Set that contains keywords about a protocol."""

    default_binary_encoding = None
    """Default encoding for binary data. It could be e.g. base64."""

    default_string_encoding = None
    """Default encoding for text content. It could be e.g. UTF-8."""

    type_attrs = {}
    """Default customizations to be passed to underlying classes."""

    def __init__(self, app=None, mime_type=None, ignore_wrappers=None,
                                    binary_encoding=None, string_encoding=None):
        self.__app = None
        self.set_app(app)

        self.ignore_wrappers = ignore_wrappers
        self.event_manager = EventManager(self)
        self.binary_encoding = binary_encoding
        if self.binary_encoding is None:
            self.binary_encoding = self.default_binary_encoding

        self.string_encoding = string_encoding
        if self.string_encoding is None:
            self.string_encoding = self.default_string_encoding

        if mime_type is not None:
            self.mime_type = mime_type

        self._attrcache = WeakKeyDictionary()
        self._sortcache = WeakKeyDictionary()

    def _cast(self, cls_attrs, inst):
        if cls_attrs.parser is not None:
            return cls_attrs.parser(inst)
        return inst

    _parse = _cast

    def _sanitize(self, cls_attrs, inst):
        if cls_attrs.sanitizer is not None:
            return cls_attrs.sanitizer(inst)
        return inst

    def _datetime_from_sec(self, cls, value):
        try:
            return datetime.fromtimestamp(value)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def _datetime_from_sec_float(self, cls, value):
        try:
            return datetime.fromtimestamp(value)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def _datetime_from_msec(self, cls, value):
        try:
            return datetime.fromtimestamp(value // 1000)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def _datetime_from_msec_float(self, cls, value):
        try:
            return datetime.fromtimestamp(value / 1000)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def _datetime_from_usec(self, cls, value):
        try:
            return datetime.fromtimestamp(value / 1e6)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def _get_datetime_format(self, cls_attrs):
        # FIXME: this should be dt_format, all other aliases are to be
        # deprecated
        dt_format = cls_attrs.datetime_format
        if dt_format is None:
            dt_format = cls_attrs.dt_format
        if dt_format is None:
            dt_format = cls_attrs.date_format
        if dt_format is None:
            dt_format = cls_attrs.out_format
        if dt_format is None:
            dt_format = cls_attrs.format

        return dt_format

    def _get_date_format(self, cls_attrs):
        date_format = cls_attrs.date_format
        if date_format is None:
            date_format = cls_attrs.format

        return date_format

    def _get_time_format(self, cls_attrs):
        time_format = cls_attrs.time_format
        if time_format is None:
            time_format = cls_attrs.format

        return time_format

    @property
    def app(self):
        return self.__app

    @staticmethod
    def strip_wrappers(cls, inst):
        ti = getattr(cls, '_type_info', {})

        while len(ti) == 1 and cls.Attributes._wrapper:
            # Wrappers are auto-generated objects that have exactly one
            # child type.
            key, = ti.keys()
            if not issubclass(cls, Array):
                inst = getattr(inst, key, None)
            cls, = ti.values()
            ti = getattr(cls, '_type_info', {})

        return cls, inst

    def set_app(self, value):
        assert self.__app is None, "One protocol instance should belong to one " \
                                   "application instance. It currently belongs " \
                                   "to: %r" % self.__app
        self.__app = value

    @staticmethod
    def issubclass(sub, cls):
        suborig = getattr(sub, '__orig__', None)
        clsorig = getattr(cls, '__orig__', None)
        return issubclass(sub if suborig is None else suborig,
                          cls if clsorig is None else clsorig)

    def get_cls_attrs(self, cls):
        logger.debug("%r attrcache size: %d", self, len(self._attrcache))
        attr = self._attrcache.get(cls, None)
        if attr is not None:
            return attr

        self._attrcache[cls] = attr = DefaultAttrDict([
                (k, getattr(cls.Attributes, k))
                        for k in dir(cls.Attributes) + META_ATTR
                                                     if not k.startswith('__')])

        if cls.Attributes.prot_attrs:
            cls_attrs = cls.Attributes.prot_attrs.get(self.__class__, {})
            # logger.debug("%r cls attr %r", cls, cls_attrs)
            attr.update(cls_attrs)

            inst_attrs = cls.Attributes.prot_attrs.get(self, {})
            # logger.debug("%r inst attr %r", cls, cls_attrs)
            attr.update(inst_attrs)

        return attr

    def get_context(self, parent, transport):
        return ProtocolContext(parent, transport)

    def generate_method_contexts(self, ctx):
        """Generates MethodContext instances for every callable assigned to the
        given method handle.

        The first element in the returned list is always the primary method
        context whereas the rest are all auxiliary method contexts.
        """

        call_handles = self.get_call_handles(ctx)
        if len(call_handles) == 0:
            raise ResourceNotFoundError(ctx.method_request_string)

        retval = []
        for d in call_handles:
            assert d is not None

            c = ctx.copy()
            c.descriptor = d

            retval.append(c)

        return retval

    def get_call_handles(self, ctx):
        """Method to be overriden to perform any sort of custom method mapping
        using any data in the method context. Returns a list of contexts.
        Can return multiple contexts if a method_request_string matches more
        than one function. (This is called the fanout mode.)
        """

        name = ctx.method_request_string
        if not name.startswith(u"{"):
            name = u'{%s}%s' % (self.app.interface.get_tns(), name)

        call_handles = self.app.interface.service_method_map.get(name, [])

        return call_handles

    def get_polymorphic_target(self, cls, inst):
        """If the protocol is polymorphic, extract what's returned by the user
        code.
        """

        if not self.polymorphic:
            logger.debug("PMORPH Skipped: %r is NOT polymorphic", self)
            return cls, False

        orig_cls = cls.__orig__ or cls

        if inst.__class__ is orig_cls:
            logger.debug("PMORPH Skipped: Instance class %r is the same as "
                                        "designated base class", inst.__class__)
            return cls, False

        if not isinstance(inst, orig_cls):
            logger.debug("PMORPH Skipped: Instance class %r is not a subclass "
                        "of designated base class %r", inst.__class__, orig_cls)
            return cls, False

        cls_attr = self.get_cls_attrs(cls)
        polymap_cls = cls_attr.polymap.get(inst.__class__, None)

        if polymap_cls is not None:
            logger.debug("PMORPH OK: cls switch with polymap: %r => %r",
                                                               cls, polymap_cls)
            return polymap_cls, True

        else:
            logger.debug("PMORPH OK: cls switch without polymap: %r => %r",
                                                            cls, inst.__class__)
            return inst.__class__, True

    @staticmethod
    def trc_verbose(cls, locale, default):
        """Translate a class.

        :param cls: class
        :param locale: locale string
        :param default: default string if no translation found
        :returns: translated string
        """

        if locale is None:
            locale = DEFAULT_LOCALE
            _log_locale = "default locale '%s'"
        else:
            _log_locale = "given locale '%s'"

        if cls.Attributes.translations is None:
            retval = default
            _log_tr = "translated to '%s' without any translations at all with"

        else:
            retval = cls.Attributes.translations.get(locale, _MISSING)
            if retval is _MISSING:
                retval = default
                _log_tr = "translated to '%s': No translation for"
            else:
                _log_tr = "translated to '%s' with"

        logger.debug(' '.join(("%r ", _log_tr, _log_locale)),
                                                            cls, retval, locale)

        return retval

    @staticmethod
    def trc(cls, locale, default):
        """Translate a class.

        :param cls: class
        :param locale: locale string
        :param default: default string if no translation found
        :returns: translated string
        """

        if locale is None:
            locale = DEFAULT_LOCALE
        if cls.Attributes.translations is not None:
            return cls.Attributes.translations.get(locale, default)
        return default

    @staticmethod
    def trd_verbose(trdict, locale, default):
        """Translate from a translations dict.

        :param trdict: translation dict
        :param locale: locale string
        :param default: default string if no translation found
        :returns: translated string
        """

        if locale is None:
            locale = DEFAULT_LOCALE
            _log_locale = "default locale '%s'"
        else:
            _log_locale = "given locale '%s'"

        if trdict is None:
            retval = default
            _log_tr = "translated to '%s' without any translations at all with"

        elif isinstance(trdict, string_types):
            retval = trdict
            _log_tr = "translated to '%s' regardless of"

        else:
            retval = trdict.get(locale, _MISSING)
            if retval is _MISSING:
                retval = default
                _log_tr = "translated to '%s': No translation for"
            else:
                _log_tr = "translated to '%s' with"

        logger.debug(' '.join(("%r ", _log_tr, _log_locale)),
                                                         trdict, retval, locale)

        return retval

    @staticmethod
    def trd(trdict, locale, default):
        """Translate from a translations dict.

        :param trdict: translation dict
        :param locale: locale string
        :param default: default string if no translation found
        :returns: translated string
        """

        if locale is None:
            locale = DEFAULT_LOCALE
        if trdict is None:
            return default
        if isinstance(trdict, string_types):
            return trdict

        return trdict.get(locale, default)

    def sort_fields(self, cls=None, items=None):
        logger.debug("%r sortcache size: %d", self, len(self._sortcache))
        retval = self._sortcache.get(cls, None)
        if retval is not None:
            return retval

        if items is None:
            items = list(cls.get_flat_type_info(cls).items())

        indexes = {}
        for k, v in items:
            order = self.get_cls_attrs(v).order
            if order is not None:
                if order < 0:
                    indexes[k] = len(items) + order
                else:
                    indexes[k] = order

        for k, v in items:
            order = self.get_cls_attrs(v).order
            if order is None:
                indexes[k] = len(indexes)

        items.sort(key=lambda x: indexes[x[0]])
        self._sortcache[cls] = items

        return items


META_ATTR = ['nullable', 'default_factory']



#!/usr/bin/env python
# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import logging
import random
import sys

# bcrypt seems to be among the latest consensus around cryptograpic circles on
# storing passwords.
# You need the package from http://code.google.com/p/py-bcrypt/
# You can install it by running easy_install py-bcrypt.
try:
    import bcrypt
except ImportError:
    print('easy_install --user py-bcrypt to get it.')
    raise

from spyne.application import Application
from spyne.decorator import rpc
from spyne.error import ArgumentError
from spyne.model.complex import ComplexModel
from spyne.model.fault import Fault
from spyne.model.primitive import Mandatory
from spyne.model.primitive import String
from spyne.protocol.soap import Soap11
from spyne.server.wsgi import WsgiApplication
from spyne.service import Service


class PublicKeyError(Fault):
    __namespace__ = 'spyne.examples.authentication'

    def __init__(self, value):
        super(PublicKeyError, self).__init__(
                                       faultstring='Value %r not found' % value)


class AuthenticationError(Fault):
    __namespace__ = 'spyne.examples.authentication'

    def __init__(self, user_name):
        # TODO: self.transport.http.resp_code = HTTP_401

        super(AuthenticationError, self).__init__(
                faultcode='Client.AuthenticationError',
                faultstring='Invalid authentication request for %r' % user_name)


class AuthorizationError(Fault):
    __namespace__ = 'spyne.examples.authentication'

    def __init__(self):
        # TODO: self.transport.http.resp_code = HTTP_401

        super(AuthorizationError, self).__init__(
                   faultcode='Client.AuthorizationError',
                   faultstring='You are not authozied to access this resource.')


class SpyneDict(dict):
    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            raise PublicKeyError(key)


class RequestHeader(ComplexModel):
    __namespace__ = 'spyne.examples.authentication'

    session_id = Mandatory.String
    user_name = Mandatory.String


class Preferences(ComplexModel):
    __namespace__ = 'spyne.examples.authentication'

    language = String(max_len=2)
    time_zone = String


user_db = {
    'neo': bcrypt.hashpw('Wh1teR@bbit', bcrypt.gensalt()),
}

session_db = set()

preferences_db = SpyneDict({
    'neo': Preferences(language='en', time_zone='Underground/Zion'),
    'smith': Preferences(language='xx', time_zone='Matrix/Core'),
})


class AuthenticationService(Service):
    __tns__ = 'spyne.examples.authentication'

    @rpc(Mandatory.String, Mandatory.String, _returns=String,
                                                    _throws=AuthenticationError)
    def authenticate(ctx, user_name, password):
        password_hash = user_db.get(user_name, None)

        if password_hash is None:
            raise AuthenticationError(user_name)

        if bcrypt.hashpw(password, password_hash) == password_hash:
            session_id = (user_name,
                                '%x' % random.randint(1 << 124, (1 << 128) - 1))
            session_db.add(session_id)

        else:
            raise AuthenticationError(user_name)

        return session_id[1]


class UserService(Service):
    __tns__ = 'spyne.examples.authentication'
    __in_header__ = RequestHeader

    @rpc(Mandatory.String, _throws=PublicKeyError, _returns=Preferences)
    def get_preferences(ctx, user_name):
        if user_name == 'smith':
            raise AuthorizationError()

        retval = preferences_db[user_name]

        return retval


def _on_method_call(ctx):
    if ctx.in_object is None:
        raise ArgumentError("RequestHeader is null")
    if not (ctx.in_header.user_name, ctx.in_header.session_id) in session_db:
        raise AuthenticationError(ctx.in_object.user_name)


UserService.event_manager.add_listener('method_call', _on_method_call)

if __name__ == '__main__':
    from spyne.util.wsgi_wrapper import run_twisted

    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)
    logging.getLogger('twisted').setLevel(logging.DEBUG)

    application = Application([AuthenticationService, UserService],
        tns='spyne.examples.authentication',
        in_protocol=Soap11(validator='lxml'),
        out_protocol=Soap11()
    )

    twisted_apps = [
        (WsgiApplication(application), 'app'),
    ]

    sys.exit(run_twisted(twisted_apps, 8000))




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""The ``spyne.protocol`` package contains the
:class:`spyne.protocol.ProtocolBase`` abstract base class. Every protocol
implementation is a subclass of ``ProtocolBase``.
"""

from spyne.protocol._base import ProtocolMixin
from spyne.protocol._inbase import InProtocolBase
from spyne.protocol._outbase import OutProtocolBase


class ProtocolBase(InProtocolBase, OutProtocolBase):
    def __init__(self, app=None, validator=None, mime_type=None,
           ignore_uncap=False, ignore_wrappers=False, binary_encoding=None,
                                                        string_encoding='utf8'):

        InProtocolBase.__init__(self, app=app, validator=validator,
                          mime_type=mime_type, ignore_wrappers=ignore_wrappers,
                                                binary_encoding=binary_encoding)

        OutProtocolBase.__init__(self, app=app, mime_type=mime_type,
                     ignore_wrappers=ignore_wrappers, ignore_uncap=ignore_uncap,
                                                binary_encoding=binary_encoding)

        self.default_string_encoding = string_encoding
        self.ignore_empty_faultactor = True




'''
This is a simple HelloWorld example showing how the NullServer works. The
NullServer is meant to be used mainly for testing.
'''

import logging
logging.basicConfig(level=logging.INFO)

from pprint import pprint

from spyne.application import Application
from spyne.protocol.soap import Soap11
from spyne.server.null import NullServer

from spyne.decorator import rpc
from spyne.service import Service
from spyne.model.complex import Iterable
from spyne.model.primitive import Integer
from spyne.model.primitive import Unicode


class HelloWorldService(Service):
    @rpc(Unicode, Integer, _returns=Iterable(Unicode))
    def say_hello(ctx, name, times):
        for i in range(times):
            yield u'Hello, %s' % name


if __name__=='__main__':
    application = Application([HelloWorldService], 'spyne.examples.hello.soap',
                in_protocol=Soap11(validator='lxml'),
                out_protocol=Soap11(pretty_print=True),
            )

    # disables context markers. set logging level to logging.INFO to enable
    # them.
    logging.getLogger('spyne.server.null').setLevel(logging.CRITICAL)

    print("With serialization")
    print("==================")
    print()

    null = NullServer(application, ostr=True)
    ret_stream = null.service.say_hello('Dave', 5)
    ret_string = ''.join(i.decode() for i in ret_stream)
    print(ret_string)
    print()

    print("Without serialization")
    print("=====================")
    print()

    null = NullServer(application, ostr=False)
    ret = null.service.say_hello('Dave', 5)
    # because the return value is a generator, we need to iterate over it to
    # see the actual return values.
    pprint(list(ret))



#!/usr/bin/env python
# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


"""As the Xml Schema standard does not define a file primitive, the File type
in the HTTP example does not work with Soap11 protocol. This is how you should
handle binary data with Soap in Spyne.

There is MTOM code inside Spyne but it lacks tests so it is not working as of
now.
"""


import logging
logger = logging.getLogger(__name__)
import os

from werkzeug.wsgi import DispatcherMiddleware
from werkzeug.exceptions import NotFound
from werkzeug.serving import run_simple

from spyne.application import Application
from spyne.decorator import rpc
from spyne.service import Service
from spyne.error import ResourceNotFoundError
from spyne.error import ValidationError
from spyne.model.binary import ByteArray
from spyne.model.primitive import Unicode
from spyne.model.primitive import Mandatory
from spyne.server.wsgi import WsgiApplication
from spyne.protocol.soap import Soap11


BLOCK_SIZE = 8192
port = 9000


class FileServices(Service):
    @rpc(Mandatory.Unicode, _returns=ByteArray(encoding='hex'))
    def get(ctx, file_name):
        path = os.path.join(os.path.abspath('./files'), file_name)
        if not path.startswith(os.path.abspath('./files')):
            raise ValidationError(file_name)

        try:
            f = open(path, 'rb')
        except IOError:
            raise ResourceNotFoundError(file_name)

        ctx.transport.resp_headers['Content-Disposition'] = (
                                         'attachment; filename=%s;' % file_name)

        data = f.read(BLOCK_SIZE)
        while len(data) > 0:
            yield data

            data = f.read(BLOCK_SIZE)

        f.close()

    @rpc(Unicode, Unicode, Unicode, ByteArray(min_occurs=1, nullable=False))
    def add(ctx, person_type, action, file_name, file_data):
        logger.info("Person Type: %r" % person_type)
        logger.info("Action: %r" % action)

        path = os.path.join(os.path.abspath('./files'), file_name)
        if not path.startswith(os.path.abspath('./files')):
            raise ValidationError(file_name)

        f = open(path, 'wb') # if this fails, the client will see an
                            # internal error.

        try:
            for data in file_data:
                f.write(data)

            logger.debug("File written: %r" % file_name)

            f.close()

        except:
            f.close()
            os.remove(path)
            logger.debug("File removed: %r" % file_name)
            raise # again, the client will see an internal error.


def main():
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)

    filemgr_app = WsgiApplication(Application([FileServices],
            tns='spyne.examples.file_manager',
            in_protocol=Soap11(validator='lxml'),
            out_protocol=Soap11()
        ))

    try:
        os.makedirs('./files')
    except OSError:
        pass

    wsgi_app = DispatcherMiddleware(NotFound(), {'/filemgr': filemgr_app})

    return run_simple('localhost', port, wsgi_app, static_files={'/': 'static'},
                                                                  threaded=True)


if __name__ == '__main__':
    import sys
    sys.exit(main())




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import print_function, unicode_literals

import logging
logger = logging.getLogger(__name__)

import re
import uuid
import errno

from os.path import isabs, join, abspath
from collections import deque
from datetime import datetime
from decimal import Decimal as D
from mmap import mmap, ACCESS_READ
from time import mktime, strftime

try:
    from lxml import etree
    from lxml import html
except ImportError:
    etree = None
    html = None

from spyne.protocol._base import ProtocolMixin
from spyne.model import ModelBase, XmlAttribute, SimpleModel, Null, \
    ByteArray, File, ComplexModelBase, AnyXml, AnyHtml, Unicode, Decimal, \
    Double, Integer, Time, DateTime, Uuid, Duration, Boolean, AnyDict, \
    AnyUri, PushBase, Date
from spyne.model.relational import FileData

from spyne.const.http import HTTP_400, HTTP_401, HTTP_404, HTTP_405, HTTP_413, \
    HTTP_500
from spyne.error import Fault, InternalError, ResourceNotFoundError, \
    RequestTooLongError, RequestNotAllowed, InvalidCredentialsError
from spyne.model.binary import binary_encoding_handlers, \
    BINARY_ENCODING_USE_DEFAULT

from spyne.util import six
from spyne.util.cdict import cdict


class OutProtocolBase(ProtocolMixin):
    """This is the abstract base class for all out protocol implementations.
    Child classes can implement only the required subset of the public methods.

    An output protocol must implement :func:`serialize` and
    :func:`create_out_string`.

    The OutProtocolBase class supports the following events:

    * ``before_serialize``:
      Called before after the serialization operation is attempted.

    * ``after_serialize``:
      Called after the serialization operation is finished.

    The arguments the constructor takes are as follows:

    :param app: The application this protocol belongs to.
    :param mime_type: The mime_type this protocol should set for transports
        that support this. This is a quick way to override the mime_type by
        default instead of subclassing the releavant protocol implementation.
    :param ignore_uncap: Silently ignore cases when the protocol is not capable
        of serializing return values instead of raising a TypeError.
    """

    def __init__(self, app=None, mime_type=None, ignore_uncap=False,
             ignore_wrappers=False, binary_encoding=None, string_encoding=None):

        super(OutProtocolBase, self).__init__(app=app, mime_type=mime_type,
            ignore_wrappers=ignore_wrappers,
               binary_encoding=binary_encoding, string_encoding=string_encoding)

        self.ignore_uncap = ignore_uncap
        self.message = None

        if mime_type is not None:
            self.mime_type = mime_type

        self._to_bytes_handlers = cdict({
            ModelBase: self.model_base_to_bytes,
            File: self.file_to_bytes,
            Time: self.time_to_bytes,
            Uuid: self.uuid_to_bytes,
            Null: self.null_to_bytes,
            Date: self.date_to_bytes,
            Double: self.double_to_bytes,
            AnyXml: self.any_xml_to_bytes,
            Unicode: self.unicode_to_bytes,
            Boolean: self.boolean_to_bytes,
            Decimal: self.decimal_to_bytes,
            Integer: self.integer_to_bytes,
            AnyHtml: self.any_html_to_bytes,
            DateTime: self.datetime_to_bytes,
            Duration: self.duration_to_bytes,
            ByteArray: self.byte_array_to_bytes,
            XmlAttribute: self.xmlattribute_to_bytes,
            ComplexModelBase: self.complex_model_base_to_bytes,
        })

        self._to_unicode_handlers = cdict({
            ModelBase: self.model_base_to_unicode,
            File: self.file_to_unicode,
            Time: self.time_to_unicode,
            Date: self.date_to_unicode,
            Uuid: self.uuid_to_unicode,
            Null: self.null_to_unicode,
            Double: self.double_to_unicode,
            AnyXml: self.any_xml_to_unicode,
            AnyUri: self.any_uri_to_unicode,
            AnyDict: self.any_dict_to_unicode,
            AnyHtml: self.any_html_to_unicode,
            Unicode: self.unicode_to_unicode,
            Boolean: self.boolean_to_unicode,
            Decimal: self.decimal_to_unicode,
            Integer: self.integer_to_unicode,
            # FIXME: Would we need a to_unicode for localized dates?
            DateTime: self.datetime_to_unicode,
            Duration: self.duration_to_unicode,
            ByteArray: self.byte_array_to_unicode,
            XmlAttribute: self.xmlattribute_to_unicode,
            ComplexModelBase: self.complex_model_base_to_unicode,
        })

        self._to_bytes_iterable_handlers = cdict({
            File: self.file_to_bytes_iterable,
            ByteArray: self.byte_array_to_bytes_iterable,
            ModelBase: self.model_base_to_bytes_iterable,
            SimpleModel: self.simple_model_to_bytes_iterable,
            ComplexModelBase: self.complex_model_to_bytes_iterable,
        })


    def serialize(self, ctx, message):
        """Serializes ``ctx.out_object``.

        If ctx.out_stream is not None,  ``ctx.out_document`` and
        ``ctx.out_string`` are skipped and the response is written directly to
        ``ctx.out_stream``.

        :param ctx: :class:`MethodContext` instance.
        :param message: One of ``(ProtocolBase.REQUEST, ProtocolBase.RESPONSE)``.
        """

    def create_out_string(self, ctx, out_string_encoding=None):
        """Uses ctx.out_document to set ctx.out_string"""

    def fault_to_http_response_code(self, fault):
        """Special function to convert native Python exceptions to Http response
        codes.
        """

        if isinstance(fault, RequestTooLongError):
            return HTTP_413

        if isinstance(fault, ResourceNotFoundError):
            return HTTP_404

        if isinstance(fault, RequestNotAllowed):
            return HTTP_405

        if isinstance(fault, InvalidCredentialsError):
            return HTTP_401

        if isinstance(fault, Fault) and (fault.faultcode.startswith('Client.')
                                                or fault.faultcode == 'Client'):
            return HTTP_400

        return HTTP_500

    def set_validator(self, validator):
        """You must override this function if you want your protocol to support
        validation."""

        assert validator is None

        self.validator = None

    def to_bytes(self, cls, value, *args, **kwargs):
        if value is None:
            return None

        handler = self._to_bytes_handlers[cls]
        retval = handler(cls, value, *args, **kwargs)

        # enable this only for testing. we're not as strict for performance
        # reasons
        # assert isinstance(retval, six.binary_type), \
        #                 "AssertionError: %r %r %r handler: %r" % \
        #                       (type(retval), six.binary_type, retval, handler)
        return retval

    def to_unicode(self, cls, value, *args, **kwargs):
        if value is None:
            return None

        handler = self._to_unicode_handlers[cls]
        retval = handler(cls, value, *args, **kwargs)

        # enable this only for testing. we're not as strict for performance
        # reasons as well as not to take the joy of dealing with duck typing
        # from the user
        # assert isinstance(retval, six.text_type), \
        #                  "AssertionError: %r %r handler: %r" % \
        #                                        (type(retval), retval, handler)

        return retval

    def to_bytes_iterable(self, cls, value):
        if value is None:
            return []

        if isinstance(value, PushBase):
            return value

        handler = self._to_bytes_iterable_handlers[cls]
        return handler(cls, value)

    def null_to_bytes(self, cls, value, **_):
        return b""

    def null_to_unicode(self, cls, value, **_):
        return u""

    def any_xml_to_bytes(self, cls, value, **_):
        return etree.tostring(value)

    def any_xml_to_unicode(self, cls, value, **_):
        return etree.tostring(value, encoding='unicode')

    def any_dict_to_unicode(self, cls, value, **_):
        return repr(value)

    def any_html_to_bytes(self, cls, value, **_):
        return html.tostring(value)

    def any_html_to_unicode(self, cls, value, **_):
        return html.tostring(value, encoding='unicode')

    def uuid_to_bytes(self, cls, value, suggested_encoding=None, **_):
        ser_as = self.get_cls_attrs(cls).serialize_as
        retval = self.uuid_to_unicode(cls, value,
                                     suggested_encoding=suggested_encoding, **_)

        if ser_as in ('bytes', 'bytes_le', 'fields', 'int', six.binary_type):
            return retval

        return retval.encode('ascii')

    def uuid_to_unicode(self, cls, value, suggested_encoding=None, **_):
        attr = self.get_cls_attrs(cls)
        ser_as = attr.serialize_as
        encoding = attr.encoding

        if encoding is None:
            encoding = suggested_encoding

        retval = _uuid_serialize[ser_as](value)
        if ser_as in ('bytes', 'bytes_le'):
            retval = binary_encoding_handlers[encoding]((retval,))
        return retval

    def unicode_to_bytes(self, cls, value, **_):
        retval = value

        cls_attrs = self.get_cls_attrs(cls)

        if isinstance(value, six.text_type):
            if cls_attrs.encoding is not None:
                retval = value.encode(cls_attrs.encoding)
            elif self.default_string_encoding is not None:
                retval = value.encode(self.default_string_encoding)
            elif not six.PY2:
                logger.warning("You need to set either an encoding for %r "
                               "or a default_string_encoding for %r", cls, self)

        if cls_attrs.str_format is not None:
            return cls_attrs.str_format.format(value)
        elif cls_attrs.format is not None:
            return cls_attrs.format % retval

        return retval

    def any_uri_to_unicode(self, cls, value, **_):
        return self.unicode_to_unicode(cls, value, **_)

    def unicode_to_unicode(self, cls, value, **_):  # :)))
        cls_attrs = self.get_cls_attrs(cls)

        retval = value

        if isinstance(value, six.binary_type):
            if cls_attrs.encoding is not None:
                retval = value.decode(cls_attrs.encoding)

            if self.default_string_encoding is not None:
                retval = value.decode(self.default_string_encoding)

            elif not six.PY2:
                logger.warning("You need to set either an encoding for %r "
                               "or a default_string_encoding for %r", cls, self)

        if cls_attrs.str_format is not None:
            return cls_attrs.str_format.format(value)
        elif cls_attrs.format is not None:
            return cls_attrs.format % retval

        return retval

    def decimal_to_bytes(self, cls, value, **_):
        return self.decimal_to_unicode(cls, value, **_).encode('utf8')

    def decimal_to_unicode(self, cls, value, **_):
        D(value)  # sanity check
        cls_attrs = self.get_cls_attrs(cls)

        if cls_attrs.str_format is not None:
            return cls_attrs.str_format.format(value)
        elif cls_attrs.format is not None:
            return cls_attrs.format % value

        return str(value)

    def double_to_bytes(self, cls, value, **_):
        return self.double_to_unicode(cls, value, **_).encode('utf8')

    def double_to_unicode(self, cls, value, **_):
        float(value) # sanity check
        cls_attrs = self.get_cls_attrs(cls)

        if cls_attrs.str_format is not None:
            return cls_attrs.str_format.format(value)
        elif cls_attrs.format is not None:
            return cls_attrs.format % value

        return repr(value)

    def integer_to_bytes(self, cls, value, **_):
        return self.integer_to_unicode(cls, value, **_).encode('utf8')

    def integer_to_unicode(self, cls, value, **_):
        int(value)  # sanity check
        cls_attrs = self.get_cls_attrs(cls)

        if cls_attrs.str_format is not None:
            return cls_attrs.str_format.format(value)
        elif cls_attrs.format is not None:
            return cls_attrs.format % value

        return str(value)

    def time_to_bytes(self, cls, value, **kwargs):
        return self.time_to_unicode(cls, value, **kwargs)

    def time_to_unicode(self, cls, value, **_):
        """Returns ISO formatted times."""
        if isinstance(value, datetime):
            value = value.time()
        return value.isoformat()

    def date_to_bytes(self, cls, val, **_):
        return self.date_to_unicode(cls, val, **_).encode("utf8")

    def date_to_unicode(self, cls, val, **_):
        if isinstance(val, datetime):
            val = val.date()

        sa = self.get_cls_attrs(cls).serialize_as

        if sa is None or sa in (str, 'str'):
            return self._date_to_bytes(cls, val)

        return _datetime_smap[sa](cls, val)

    def datetime_to_bytes(self, cls, val, **_):
        retval = self.datetime_to_unicode(cls, val, **_)
        sa = self.get_cls_attrs(cls).serialize_as
        if sa is None or sa in (six.text_type, str, 'str'):
            return retval.encode('ascii')
        return retval

    def datetime_to_unicode(self, cls, val, **_):
        sa = self.get_cls_attrs(cls).serialize_as

        if sa is None or sa in (six.text_type, str, 'str'):
            return self._datetime_to_unicode(cls, val)

        return _datetime_smap[sa](cls, val)

    def duration_to_bytes(self, cls, value, **_):
        return self.duration_to_unicode(cls, value, **_).encode("utf8")

    def duration_to_unicode(self, cls, value, **_):
        if value.days < 0:
            value = -value
            negative = True
        else:
            negative = False

        tot_sec = int(value.total_seconds())
        seconds = value.seconds % 60
        minutes = value.seconds // 60
        hours = minutes // 60
        minutes %= 60
        seconds = float(seconds)
        useconds = value.microseconds

        retval = deque()
        if negative:
            retval.append("-P")
        else:
            retval.append("P")
        if value.days != 0:
            retval.append("%iD" % value.days)

        if tot_sec != 0 and tot_sec % 86400 == 0 and useconds == 0:
            return ''.join(retval)

        retval.append('T')

        if hours > 0:
            retval.append("%iH" % hours)

        if minutes > 0:
            retval.append("%iM" % minutes)

        if seconds > 0 or useconds > 0:
            retval.append("%i" % seconds)
            if useconds > 0:
                retval.append(".%i" % useconds)
            retval.append("S")

        if len(retval) == 2:
            retval.append('0S')

        return ''.join(retval)

    def boolean_to_bytes(self, cls, value, **_):
        return str(bool(value)).lower().encode('ascii')

    def boolean_to_unicode(self, cls, value, **_):
        return str(bool(value)).lower()

    def byte_array_to_bytes(self, cls, value, suggested_encoding=None, **_):
        cls_attrs = self.get_cls_attrs(cls)

        encoding = cls_attrs.encoding
        if encoding is BINARY_ENCODING_USE_DEFAULT:
            if suggested_encoding is None:
                encoding = self.binary_encoding
            else:
                encoding = suggested_encoding

        if encoding is None and isinstance(value, (list, tuple)) \
                             and len(value) == 1 and isinstance(value[0], mmap):
            return value[0]

        encoder = binary_encoding_handlers[encoding]
        logger.debug("Using binary encoder %r for encoding %r",
                                                              encoder, encoding)
        retval = encoder(value)
        if encoding is not None and isinstance(retval, six.text_type):
            retval = retval.encode('ascii')

        return retval

    def byte_array_to_unicode(self, cls, value, suggested_encoding=None, **_):
        encoding = self.get_cls_attrs(cls).encoding
        if encoding is BINARY_ENCODING_USE_DEFAULT:
            if suggested_encoding is None:
                encoding = self.binary_encoding
            else:
                encoding = suggested_encoding

        if encoding is None:
            raise ValueError("Arbitrary binary data can't be serialized to "
                                                                      "unicode")

        retval = binary_encoding_handlers[encoding](value)
        if not isinstance(retval, six.text_type):
            retval = retval.decode('ascii')

        return retval

    def byte_array_to_bytes_iterable(self, cls, value, **_):
        return value

    def file_to_bytes(self, cls, value, suggested_encoding=None):
        """
        :param cls: A :class:`spyne.model.File` subclass
        :param value: Either a sequence of byte chunks or a
            :class:`spyne.model.File.Value` instance.
        """

        encoding = self.get_cls_attrs(cls).encoding
        if encoding is BINARY_ENCODING_USE_DEFAULT:
            if suggested_encoding is None:
                encoding = self.binary_encoding
            else:
                encoding = suggested_encoding

        if isinstance(value, File.Value):
            if value.data is not None:
                return binary_encoding_handlers[encoding](value.data)

            if value.handle is not None:
                # maybe we should have used the sweeping except: here.
                if hasattr(value.handle, 'fileno'):
                    if six.PY2:
                        fileno = value.handle.fileno()
                        data = (mmap(fileno, 0, access=ACCESS_READ),)
                    else:
                        import io
                        try:
                            fileno = value.handle.fileno()
                            data = mmap(fileno, 0, access=ACCESS_READ)
                        except io.UnsupportedOperation:
                            data = (value.handle.read(),)
                else:
                    data = (value.handle.read(),)

                return binary_encoding_handlers[encoding](data)

            if value.path is not None:
                handle = open(value.path, 'rb')
                fileno = handle.fileno()
                data = mmap(fileno, 0, access=ACCESS_READ)

                return binary_encoding_handlers[encoding](data)

            assert False, "Unhandled file type"

        if isinstance(value, FileData):
            try:
                return binary_encoding_handlers[encoding](value.data)
            except Exception as e:
                logger.error("Error encoding value to binary. Error: %r, Value: %r",
                                                                           e, value)
                raise

        try:
            return binary_encoding_handlers[encoding](value)
        except Exception as e:
            logger.error("Error encoding value to binary. Error: %r, Value: %r",
                                                                       e, value)
            raise

    def file_to_unicode(self, cls, value, suggested_encoding=None):
        """
        :param cls: A :class:`spyne.model.File` subclass
        :param value: Either a sequence of byte chunks or a
            :class:`spyne.model.File.Value` instance.
        """

        cls_attrs = self.get_cls_attrs(cls)
        encoding = cls_attrs.encoding
        if encoding is BINARY_ENCODING_USE_DEFAULT:
            encoding = suggested_encoding

        if encoding is None and cls_attrs.mode is File.TEXT:
            raise ValueError("Arbitrary binary data can't be serialized to "
                             "unicode.")

        retval = self.file_to_bytes(cls, value, suggested_encoding)
        if not isinstance(retval, six.text_type):
            retval = retval.decode('ascii')
        return retval

    def file_to_bytes_iterable(self, cls, value, **_):
        if value.data is not None:
            if isinstance(value.data, (list, tuple)) and \
                                                isinstance(value.data[0], mmap):
                return _file_to_iter(value.data[0])
            return iter(value.data)

        if value.handle is not None:
            f = value.handle
            f.seek(0)
            return _file_to_iter(f)

        assert value.path is not None, "You need to write data to " \
                 "persistent storage first if you want to read it back."

        try:
            path = value.path
            if not isabs(value.path):
                path = join(value.store, value.path)
                assert abspath(path).startswith(value.store), \
                                                 "No relative paths are allowed"
            return _file_to_iter(open(path, 'rb'))

        except IOError as e:
            if e.errno == errno.ENOENT:
                raise ResourceNotFoundError(value.path)
            else:
                raise InternalError("Error accessing requested file")

    def simple_model_to_bytes_iterable(self, cls, value, **kwargs):
        retval = self.to_bytes(cls, value, **kwargs)
        if retval is None:
            return (b'',)
        return (retval,)

    def complex_model_to_bytes_iterable(self, cls, value, **_):
        if self.ignore_uncap:
            return tuple()
        raise TypeError("This protocol can only serialize primitives.")

    def complex_model_base_to_bytes(self, cls, value, **_):
        raise TypeError("Only primitives can be serialized to string.")

    def complex_model_base_to_unicode(self, cls, value, **_):
        raise TypeError("Only primitives can be serialized to string.")

    def xmlattribute_to_bytes(self, cls, string, **kwargs):
        return self.to_bytes(cls.type, string, **kwargs)

    def xmlattribute_to_unicode(self, cls, string, **kwargs):
        return self.to_unicode(cls.type, string, **kwargs)

    def model_base_to_bytes_iterable(self, cls, value, **kwargs):
        return cls.to_bytes_iterable(value, **kwargs)

    def model_base_to_bytes(self, cls, value, **kwargs):
        return cls.to_bytes(value, **kwargs)

    def model_base_to_unicode(self, cls, value, **kwargs):
        return cls.to_unicode(value, **kwargs)

    def _datetime_to_unicode(self, cls, value, **_):
        """Returns ISO formatted datetimes."""

        cls_attrs = self.get_cls_attrs(cls)

        if cls_attrs.as_timezone is not None and value.tzinfo is not None:
            value = value.astimezone(cls_attrs.as_timezone)

        if not cls_attrs.timezone:
            value = value.replace(tzinfo=None)

        dt_format = self._get_datetime_format(cls_attrs)

        if dt_format is None:
            retval = value.isoformat()

        elif six.PY2 and isinstance(dt_format, unicode):
            retval = self.strftime(value, dt_format.encode('utf8')).decode('utf8')

        else:
            retval = self.strftime(value, dt_format)

        # FIXME: must deprecate string_format, this should have been str_format
        str_format = cls_attrs.string_format
        if str_format is None:
            str_format = cls_attrs.str_format
        if str_format is not None:
            return str_format.format(value)

        # FIXME: must deprecate interp_format, this should have been just format
        interp_format = cls_attrs.interp_format
        if interp_format is not None:
            return interp_format.format(value)

        return retval

    def _date_to_bytes(self, cls, value, **_):
        cls_attrs = self.get_cls_attrs(cls)

        date_format = cls_attrs.date_format
        if date_format is None:
            retval = value.isoformat()

        elif six.PY2 and isinstance(date_format, unicode):
            date_format = date_format.encode('utf8')
            retval = self.strftime(value, date_format).decode('utf8')

        else:
            retval = self.strftime(value, date_format)

        str_format = cls_attrs.str_format
        if str_format is not None:
            return str_format.format(value)

        format = cls_attrs.format
        if format is not None:
            return format.format(value)

        return retval

    # Format a datetime through its full proleptic Gregorian date range.
    # http://code.activestate.com/recipes/
    #                306860-proleptic-gregorian-dates-and-strftime-before-1900/
    # http://stackoverflow.com/a/32206673
    #
    # >>> strftime(datetime.date(1850, 8, 2), "%Y/%M/%d was a %A")
    # '1850/00/02 was a Friday'
    # >>>


    # remove the unsupposed "%s" command.  But don't
    # do it if there's an even number of %s before the s
    # because those are all escaped.  Can't simply
    # remove the s because the result of
    #  %sY
    # should be %Y if %s isn't supported, not the
    # 4 digit year.
    _illegal_s = re.compile(r"((^|[^%])(%%)*%s)")

    @staticmethod
    def _findall_datetime(text, substr):
         # Also finds overlaps
         sites = []
         i = 0
         while 1:
             j = text.find(substr, i)
             if j == -1:
                 break
             sites.append(j)
             i=j+1
         return sites

    # Every 28 years the calendar repeats, except through century leap
    # years where it's 6 years.  But only if you're using the Gregorian
    # calendar.  ;)

    @classmethod
    def strftime(cls, dt, fmt):
        if cls._illegal_s.search(fmt):
            raise TypeError("This strftime implementation does not handle %s")
        if dt.year > 1900:
            return dt.strftime(fmt)

        year = dt.year
        # For every non-leap year century, advance by
        # 6 years to get into the 28-year repeat cycle
        delta = 2000 - year
        off = 6*(delta // 100 + delta // 400)
        year += off

        # Move to around the year 2000
        year += ((2000 - year) // 28) * 28
        timetuple = dt.timetuple()
        s1 = strftime(fmt, (year,) + timetuple[1:])
        sites1 = cls._findall_datetime(s1, str(year))

        s2 = strftime(fmt, (year+28,) + timetuple[1:])
        sites2 = cls._findall_datetime(s2, str(year+28))

        sites = []
        for site in sites1:
            if site in sites2:
                sites.append(site)

        s = s1
        syear = "%4d" % (dt.year,)
        for site in sites:
            s = s[:site] + syear + s[site+4:]
        return s


_uuid_serialize = {
    None: str,
    str: str,
    'str': str,

    'hex': lambda u: u.hex,
    'urn': lambda u: u.urn,
    'bytes': lambda u: u.bytes,
    'bytes_le': lambda u: u.bytes_le,
    'fields': lambda u: u.fields,

    int: lambda u: u.int,
    'int': lambda u: u.int,
}

_uuid_deserialize = {
    None: uuid.UUID,
    str: uuid.UUID,
    'str': uuid.UUID,

    'hex': lambda s: uuid.UUID(hex=s),
    'urn': lambda s: uuid.UUID(hex=s),
    'bytes': lambda s: uuid.UUID(bytes=s),
    'bytes_le': lambda s: uuid.UUID(bytes_le=s),
    'fields': lambda s: uuid.UUID(fields=s),

    int: lambda s: uuid.UUID(int=s),
    'int': lambda s: uuid.UUID(int=s),

    (int, int): lambda s: uuid.UUID(int=s),
    ('int', int): lambda s: uuid.UUID(int=s),

    (int, str): lambda s: uuid.UUID(int=int(s)),
    ('int', str): lambda s: uuid.UUID(int=int(s)),
}

if six.PY2:
    _uuid_deserialize[('int', long)] = _uuid_deserialize[('int', int)]
    _uuid_deserialize[(int, long)] = _uuid_deserialize[('int', int)]


def _parse_datetime_iso_match(date_match, tz=None):
    fields = date_match.groupdict()

    year = int(fields.get('year'))
    month = int(fields.get('month'))
    day = int(fields.get('day'))
    hour = int(fields.get('hr'))
    minute = int(fields.get('min'))
    second = int(fields.get('sec'))
    usecond = fields.get("sec_frac")
    if usecond is None:
        usecond = 0
    else:
        # we only get the most significant 6 digits because that's what
        # datetime can handle.
        usecond = int(round(float(usecond) * 1e6))

    return datetime(year, month, day, hour, minute, second, usecond, tz)


_dt_sec = lambda cls, val: \
        int(mktime(val.timetuple()))
_dt_sec_float = lambda cls, val: \
        mktime(val.timetuple()) + (val.microsecond / 1e6)

_dt_msec = lambda cls, val: \
        int(mktime(val.timetuple())) * 1000 + (val.microsecond // 1000)
_dt_msec_float = lambda cls, val: \
        mktime(val.timetuple()) * 1000 + (val.microsecond / 1000.0)

_dt_usec = lambda cls, val: \
        int(mktime(val.timetuple())) * 1000000 + val.microsecond

_datetime_smap = {
    'sec': _dt_sec,
    'secs': _dt_sec,
    'second': _dt_sec,
    'seconds': _dt_sec,

    'sec_float': _dt_sec_float,
    'secs_float': _dt_sec_float,
    'second_float': _dt_sec_float,
    'seconds_float': _dt_sec_float,

    'msec': _dt_msec,
    'msecs': _dt_msec,
    'msecond': _dt_msec,
    'mseconds': _dt_msec,
    'millisecond': _dt_msec,
    'milliseconds': _dt_msec,

    'msec_float': _dt_msec_float,
    'msecs_float': _dt_msec_float,
    'msecond_float': _dt_msec_float,
    'mseconds_float': _dt_msec_float,
    'millisecond_float': _dt_msec_float,
    'milliseconds_float': _dt_msec_float,

    'usec': _dt_usec,
    'usecs': _dt_usec,
    'usecond': _dt_usec,
    'useconds': _dt_usec,
    'microsecond': _dt_usec,
    'microseconds': _dt_usec,
}


def _file_to_iter(f):
    try:
        data = f.read(8192)
        while len(data) > 0:
            yield data
            data = f.read(8192)

    finally:
        f.close()


META_ATTR = ['nullable', 'default_factory']



#!/usr/bin/env python

from __future__ import unicode_literals

import unittest

from lxml import etree
from lxml.doctestcompare import LXMLOutputChecker, PARSE_XML

from spyne import Fault, Unicode, ByteArray
from spyne.application import Application
from spyne.const import xml as ns
from spyne.const.xml import NS_SOAP11_ENV
from spyne.decorator import srpc, rpc
from spyne.interface import Wsdl11
from spyne.model.complex import ComplexModel
from spyne.model.primitive import Integer, String
from spyne.protocol.soap.mime import _join_attachment
from spyne.protocol.soap.soap12 import Soap12
from spyne.protocol.xml import XmlDocument
from spyne.server.wsgi import WsgiApplication
from spyne.service import Service
from spyne.test.protocol.test_soap11 import TestService, TestSingle, \
    TestReturn, MultipleReturnService
from spyne.util.six import BytesIO


def start_response(code, headers):
    print(code, headers)


MTOM_REQUEST = b"""
--uuid:2e53e161-b47f-444a-b594-eb6b72e76997
Content-Type: application/xop+xml; charset=UTF-8;
  type="application/soap+xml"; action="sendDocument";
Content-Transfer-Encoding: binary
Content-ID: <root.message@cxf.apache.org>

<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope">
  <soap:Body>
    <ns3:documentRequest xmlns:xmime="http://www.w3.org/2005/05/xmlmime" xmlns:ns3="http://gib.gov.tr/vedop3/eFatura">
      <fileName>EA055406-5881-4F02-A3DC-9A5A7510D018.dat</fileName>
      <binaryData xmime:contentType="application/octet-stream">
        <xop:Include xmlns:xop="http://www.w3.org/2004/08/xop/include" href="cid:04dfbca1-54b8-4631-a556-4addea6716ed-223384@cxf.apache.org"/>
      </binaryData>
      <hash>26981FCD51C95FA47780400B7A45132F</hash>
    </ns3:documentRequest>
  </soap:Body>
</soap:Envelope>

--uuid:2e53e161-b47f-444a-b594-eb6b72e76997
Content-Type: application/octet-stream
Content-Transfer-Encoding: binary
Content-ID: <04dfbca1-54b8-4631-a556-4addea6716ed-223384@cxf.apache.org>

sample data
--uuid:2e53e161-b47f-444a-b594-eb6b72e76997--
"""


# Service Classes
class DownloadPartFileResult(ComplexModel):
    ErrorCode = Integer
    ErrorMessage = String
    Data = String


class TestSingleSoap12(TestSingle):
    def setUp(self):
        self.app = Application([TestService], 'tns', in_protocol=Soap12(), out_protocol=Soap12())
        self.app.transport = 'null.spyne'
        self.srv = TestService()

        wsdl = Wsdl11(self.app.interface)
        wsdl.build_interface_document('URL')
        self.wsdl_str = wsdl.get_interface_document()
        self.wsdl_doc = etree.fromstring(self.wsdl_str)


class TestMultipleSoap12(TestReturn):
    def setUp(self):
        self.app = Application([MultipleReturnService], 'tns', in_protocol=Soap12(), out_protocol=Soap12())
        self.app.transport = 'none'
        self.wsdl = Wsdl11(self.app.interface)
        self.wsdl.build_interface_document('URL')


class TestSoap12(unittest.TestCase):

    def test_soap12(self):
        element = etree.fromstring(b"""
        <soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope">
          <soap:Body>
            <soap:Fault>
                <soap:Code>
                    <soap:Value>env:Sender</soap:Value>
                    <soap:Subcode>
                        <soap:Value>st:SomeDomainProblem</soap:Value>
                    </soap:Subcode>
                </soap:Code>
                <soap:Reason>
                    <soap:Text xml:lang="en-US">
                        Some_Policy
                    </soap:Text>
                </soap:Reason>
            </soap:Fault>
          </soap:Body>
        </soap:Envelope>""")

        so = Soap12()
        ret = so.from_element(None, Fault, element[0][0])
        assert ret.faultcode == "env:Sender.st:SomeDomainProblem"

    def test_fault_generation(self):
        class SoapException(Service):
            @srpc()
            def soap_exception():
                raise Fault(
                    "Client.Plausible.issue", "A plausible fault", 'http://faultactor.example.com',
                    detail={'some':'extra info'})
        app = Application([SoapException], 'tns', in_protocol=Soap12(), out_protocol=Soap12())

        req = b"""
        <soap12env:Envelope
                xmlns:soap12env="http://www.w3.org/2003/05/soap-envelope"
                xmlns:tns="tns">
            <soap12env:Body>
                <tns:soap_exception/>
            </soap12env:Body>
        </soap12env:Envelope>
        """

        server = WsgiApplication(app)
        response = etree.fromstring(b''.join(server({
            'QUERY_STRING': '',
            'PATH_INFO': '/call',
            'REQUEST_METHOD': 'POST',
            'CONTENT_TYPE': 'text/xml; charset=utf8',
            'wsgi.input': BytesIO(req)
        }, start_response, "http://null")))

        response_str = etree.tostring(response, pretty_print=True)
        print(response_str)

        expected = b"""
            <soap12env:Envelope xmlns:soap12env="http://www.w3.org/2003/05/soap-envelope">
              <soap12env:Body>
                <soap12env:Fault>
                  <soap12env:Code>
                    <soap12env:Value>soap12env:Sender</soap12env:Value>
                    <soap12env:Subcode>
                      <soap12env:Value>Plausible</soap12env:Value>
                      <soap12env:Subcode>
                        <soap12env:Value>issue</soap12env:Value>
                      </soap12env:Subcode>
                    </soap12env:Subcode>
                  </soap12env:Code>
                  <soap12env:Reason>
                      <soap12env:Text xml:lang="en">A plausible fault</soap12env:Text>
                  </soap12env:Reason>
                  <soap12env:Role>http://faultactor.example.com</soap12env:Role>
                  <soap12env:Detail>
                    <some>extra info</some>
                  </soap12env:Detail>
                </soap12env:Fault>
              </soap12env:Body>
            </soap12env:Envelope>"""
        if not LXMLOutputChecker().check_output(expected, response_str, PARSE_XML):
            raise Exception("Got: %s but expected: %s" % (response_str, expected))

    def test_gen_fault_codes(self):
        fault_string = "Server.Plausible.error"
        value, faultstrings = Soap12().gen_fault_codes(faultstring=fault_string)
        self.assertEqual(value, "%s:Receiver" %(Soap12.soap_env))
        self.assertEqual(faultstrings[0], "Plausible")
        self.assertEqual(faultstrings[1], "error")

        fault_string = "UnknownFaultCode.Plausible.error"
        with self.assertRaises(TypeError):
            value, faultstrings = Soap12().gen_fault_codes(faultstring=fault_string)

    def test_mtom(self):
        FILE_NAME = 'EA055406-5881-4F02-A3DC-9A5A7510D018.dat'
        TNS = 'http://gib.gov.tr/vedop3/eFatura'
        class SomeService(Service):
            @rpc(Unicode(sub_name="fileName"), ByteArray(sub_name='binaryData'),
                 ByteArray(sub_name="hash"), _returns=Unicode)
            def documentRequest(ctx, file_name, file_data, data_hash):
                assert file_name == FILE_NAME
                assert file_data == (b'sample data',)

                return file_name

        app = Application([SomeService], tns=TNS,
                                    in_protocol=Soap12(), out_protocol=Soap12())

        server = WsgiApplication(app)
        response = etree.fromstring(b''.join(server({
            'QUERY_STRING': '',
            'PATH_INFO': '/call',
            'REQUEST_METHOD': 'POST',
            'CONTENT_TYPE': 'Content-Type: multipart/related; '
                            'type="application/xop+xml"; '
                            'boundary="uuid:2e53e161-b47f-444a-b594-eb6b72e76997"; '
                            'start="<root.message@cxf.apache.org>"; '
                            'start-info="application/soap+xml"; action="sendDocument"',
            'wsgi.input': BytesIO(MTOM_REQUEST.replace(b"\n", b"\r\n"))
        }, start_response, "http://null")))

        response_str = etree.tostring(response, pretty_print=True)
        print(response_str)

        nsdict = dict(tns=TNS)

        assert etree.fromstring(response_str) \
            .xpath(".//tns:documentRequestResult/text()", namespaces=nsdict) \
                                                                  == [FILE_NAME]

    def test_mtom_join_envelope_chunks(self):
        FILE_NAME = 'EA055406-5881-4F02-A3DC-9A5A7510D018.dat'
        TNS = 'http://gib.gov.tr/vedop3/eFatura'

        # large enough payload to be chunked
        PAYLOAD = b"sample data " * 1024
        class SomeService(Service):
            @rpc(Unicode(sub_name="fileName"), ByteArray(sub_name='binaryData'),
                 ByteArray(sub_name="hash"), _returns=Unicode)
            def documentRequest(ctx, file_name, file_data, data_hash):
                assert file_name == FILE_NAME
                assert file_data == (PAYLOAD,)

                return file_name

        app = Application([SomeService], tns=TNS,
                                    in_protocol=Soap12(), out_protocol=Soap12())

        server = WsgiApplication(app, block_length=1024)
        response = etree.fromstring(b''.join(server({
            'QUERY_STRING': '',
            'PATH_INFO': '/call',
            'REQUEST_METHOD': 'POST',
            'CONTENT_TYPE': 'Content-Type: multipart/related; '
                            'type="application/xop+xml"; '
                            'boundary="uuid:2e53e161-b47f-444a-b594-eb6b72e76997"; '
                            'start="<root.message@cxf.apache.org>"; '
                            'start-info="application/soap+xml"; action="sendDocument"',
            'wsgi.input': BytesIO(MTOM_REQUEST
                                  .replace(b"\n", b"\r\n")
                                  .replace(b"sample data", PAYLOAD)),
        }, start_response, "http://null")))

        response_str = etree.tostring(response, pretty_print=True)
        print(response_str)

        nsdict = dict(tns=TNS)

        assert etree.fromstring(response_str) \
            .xpath(".//tns:documentRequestResult/text()", namespaces=nsdict) \
                                                                  == [FILE_NAME]

    def test_bytes_join_attachment(self):
        href_id = "http://tempuri.org/1/634133419330914808"
        payload = "ANJNSLJNDYBC SFDJNIREMX:CMKSAJN"
        envelope = '''
<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">
  <s:Body>
    <DownloadPartFileResponse xmlns="http://tempuri.org/">
      <DownloadPartFileResult
            xmlns:a="http://schemas.datacontract.org/2004/07/KlanApi.Common"
            xmlns:i="http://www.w3.org/2001/XMLSchema-instance">
        <a:ErrorCode>0</a:ErrorCode>
        <a:ErrorMessage i:nil="true"/>
        <a:Data>
          <xop:Include href="cid:%s"
                             xmlns:xop="http://www.w3.org/2004/08/xop/include"/>
        </a:Data>
      </DownloadPartFileResult>
    </DownloadPartFileResponse>
  </s:Body>
</s:Envelope>
        ''' % href_id

        (joinedmsg, numreplaces) = _join_attachment(NS_SOAP11_ENV,
                                                     href_id, envelope, payload)

        soaptree = etree.fromstring(joinedmsg)

        body = soaptree.find(ns.SOAP11_ENV("Body"))
        response = body.getchildren()[0]
        result = response.getchildren()[0]
        r = XmlDocument().from_element(None, DownloadPartFileResult, result)

        self.assertEqual(payload, r.Data)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


"""
This example shows how to override the variable names for fun and profit.
This is very useful for situations that require the use of variable names
that are python keywords like from, import, return, etc.
"""


from spyne import Application, rpc, Service, String, ComplexModel

from spyne.util.odict import odict

from spyne.protocol.soap import Soap11
from spyne.server.wsgi import WsgiApplication


class SomeClass(ComplexModel):
    _type_info = odict([
        ('and', String),
        ('or', String),
    ])


class EmailManager(Service):
    @rpc(String, String, String, _returns=String,
        _in_variable_names={'from_': 'from'},
        _out_variable_name='return')
    def send_email(ctx, to, from_, message):
        # do email sending here
        return repr((to, from_, message, 'sent!'))


application = Application([EmailManager], 'spyne.examples.events',
                                in_protocol=Soap11(), out_protocol=Soap11())

if __name__ == '__main__':
    import logging

    from wsgiref.simple_server import make_server

    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)

    server = make_server('127.0.0.1', 8000, WsgiApplication(application))

    logging.info("listening to http://127.0.0.1:8000")
    logging.info("wsdl is at: http://localhost:8000/?wsdl")

    server.serve_forever()



# encoding: utf8
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from lxml.builder import E
from pprint import pformat

from spyne import Boolean
from spyne.protocol.html import HtmlBase


class PrettyFormat(HtmlBase):
    def to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        parent.write(E.pre(pformat(inst)))


class BooleanListProtocol(HtmlBase):
    def __init__(self, nothing=None):
        super(BooleanListProtocol, self).__init__()

        self.nothing = nothing

    def to_parent(self, ctx, cls, inst, parent, name, nosubprot=False, **kwargs):
        if inst is None:
            return

        wrote_nothing = True
        for k, v in cls.get_flat_type_info(cls).items():
            if not issubclass(v, Boolean):
                continue

            if getattr(inst, k, False):
                parent.write(E.p(self.trc(cls, ctx.locale, k)))
                wrote_nothing = False

        if wrote_nothing and self.nothing is not None:
            parent.write(E.p(self.nothing))



#!/usr/bin/env python
# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

"""
This example is an enhanced version of the HelloWorld example that
uses event listeners to apply cross-cutting behavior to the service.
In this example, the service hooks are used to gather performance
information on both the method execution as well as the duration
of the entire call, including serialization and deserialization.

Events can be used for doing things like like database transaction management,
logging and measuring performance. This example also uses the user-defined
context (udc) attribute of the MethodContext object to hold the data points for
this request.

Use:

    curl 'http://localhost:8000/say_hello?name=Dave&times=5'

to query this code.
"""

import logging

from time import time

from spyne import Application, rpc, Service, String, Integer

from spyne.server.wsgi import WsgiApplication
from spyne.protocol.json import JsonDocument
from spyne.protocol.http import HttpRpc


class UserDefinedContext(object):
    def __init__(self):
        self.call_start = time()
        self.call_end = None
        self.method_start = None
        self.method_end = None


class HelloWorldService(Service):
    @rpc(String, Integer, _returns=String(max_occurs='unbounded'))
    def say_hello(ctx, name, times):
        results = []

        for i in range(0, times):
            results.append('Hello, %s' % name)

        return results


def _on_wsgi_call(ctx):
    print("_on_wsgi_call")
    ctx.udc = UserDefinedContext()


def _on_method_call(ctx):
    print("_on_method_call")
    ctx.udc.method_start = time()


def _on_method_return_object(ctx):
    print("_on_method_return_object")
    ctx.udc.method_end = time()


def _on_wsgi_return(ctx):
    print("_on_wsgi_return")
    call_end = time()
    print('Method took [%0.8f] - total execution time[%0.8f]'% (
        ctx.udc.method_end - ctx.udc.method_start,
        call_end - ctx.udc.call_start))

def _on_wsgi_close(ctx):
    print("_on_wsgi_close: request processing completed.")

def _on_method_context_destroyed(ctx):
    print("_on_method_context_destroyed")
    print('MethodContext(%d) lived for [%0.8f] seconds' % (id(ctx),
                                                ctx.call_end - ctx.call_start))


def _on_method_context_constructed(ctx):
    print("_on_method_context_constructed")
    print('Hello, this is MethodContext(%d). Time now: %0.8f' % (id(ctx),
                                                                ctx.call_start))

if __name__=='__main__':
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)

    try:
        from wsgiref.simple_server import make_server
    except ImportError:
        logging.error("Error: example server code requires Python >= 2.5")

    application = Application([HelloWorldService], 'spyne.examples.events',
                   in_protocol=HttpRpc(), out_protocol=JsonDocument())

    application.event_manager.add_listener('method_call', _on_method_call)
    application.event_manager.add_listener('method_return_object',
                                                _on_method_return_object)
    application.event_manager.add_listener('method_context_constructed',
                                                _on_method_context_constructed)
    application.event_manager.add_listener('method_context_destroyed',
                                                _on_method_context_destroyed)

    wsgi_wrapper = WsgiApplication(application)
    wsgi_wrapper.event_manager.add_listener('wsgi_call', _on_wsgi_call)
    wsgi_wrapper.event_manager.add_listener('wsgi_return', _on_wsgi_return)
    wsgi_wrapper.event_manager.add_listener('wsgi_close', _on_wsgi_close)

    server = make_server('127.0.0.1', 8000, wsgi_wrapper)

    logging.info("listening to http://127.0.0.1:8000")
    logging.info("wsdl is at: http://localhost:8000/?wsdl")

    server.serve_forever()



#!/usr/bin/env python
# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import logging

logging.basicConfig(level=logging.DEBUG)

import uuid

from datetime import datetime
from pprint import pprint

from lxml import etree
from lxml.builder import E

from spyne import AnyXml, Uuid, Unicode, String, Integer, Decimal, DateTime, \
    XmlData, ComplexModel, XmlAttribute

from spyne.util.xml import get_schema_documents
from spyne.util.xml import get_object_as_xml
from spyne.util.xml import get_xml_as_object
from spyne.util.xml import get_validation_schema


class Punk(ComplexModel):
    __namespace__ = 'some_namespace'

    a = String
    b = Integer
    c = Decimal
    d = DateTime


class DateTimeWithAttribute(ComplexModel):
    value = XmlData(DateTime)
    f = XmlAttribute(Unicode)


class Foo(ComplexModel):
    __namespace__ = 'some_other_namespace'

    a = String
    b = Integer
    c = Decimal
    d = DateTimeWithAttribute
    e = XmlAttribute(Integer)


class ProductEdition(ComplexModel):
    __namespace__ = 'kickass_namespace'

    id = XmlAttribute(Uuid)
    name = XmlData(Unicode)
    addtl = XmlData(AnyXml)


class Product(ComplexModel):
    __namespace__ = 'kickass_namespace'

    id = XmlAttribute(Uuid)
    edition = ProductEdition


docs = get_schema_documents([Punk, Foo, Product])
pprint(docs)
print()

pprint({k: v.attrib['targetNamespace'] for k, v in docs.items()})

# the default ns prefix is always tns
print("the default namespace %r:" % docs['tns'].attrib['targetNamespace'])
print(etree.tostring(docs['tns'], pretty_print=True, encoding='unicode'))
print()

# Namespace prefixes are assigned like s0, s1, s2, etc...
print("the other namespace %r:" % docs['s0'].attrib['targetNamespace'])
print(etree.tostring(docs['s0'], pretty_print=True, encoding='unicode'))
print()

print("the other namespace %r:" % docs['s2'].attrib['targetNamespace'])
print(etree.tostring(docs['s2'], pretty_print=True, encoding='unicode'))
print()

# Object serialization and deserialization
foo = Foo(
    a='a',
    b=1,
    c=3.4,
    d=DateTimeWithAttribute(value=datetime(2011, 2, 20), f='f'),
    e=5,
)

doc = get_object_as_xml(foo, Foo)
print()
print(etree.tostring(doc, pretty_print=True, encoding='unicode'))
print(get_xml_as_object(doc, Foo))
print()

# could be anything, really
elt = E.tag(E.subtag("subdata"))

# XmlData example.
print("Product output (illustrates XmlData):")
product = Product(
    id=uuid.uuid4(),
    edition=ProductEdition(id=uuid.uuid4(), name='My edition', addtl=elt)
)

print()
print(etree.tostring(
    get_object_as_xml(product, Product),
    pretty_print=True, encoding='unicode'
))

# See http://lxml.de/validation.html to see what this could be used for.
print(get_validation_schema([Punk, Foo]))



# encoding: utf8
#
# Copyright © Anton Egorov <anton.egoroff@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


from spyne import Iterable, Integer, Unicode, rpc, Application, Service
from spyne.protocol.http import HttpRpc
from spyne.protocol.json import JsonDocument


class HelloWorldService(Service):
    @rpc(Unicode, Integer, _returns=Iterable(Unicode))
    def hello(ctx, name, times):
        name = name or ctx.udc.config['HELLO']
        for i in range(times):
            yield u'Hello, %s' % name


class UserDefinedContext(object):
    def __init__(self, flask_config):
        self.config = flask_config


def create_app(flask_app):
    """Creates SOAP services application and distribute Flask config into
    user con defined context for each method call.
    """
    application = Application(
        [HelloWorldService], 'spyne.examples.flask',
        # The input protocol is set as HttpRpc to make our service easy to call.
        in_protocol=HttpRpc(validator='soft'),
        out_protocol=JsonDocument(ignore_wrappers=True),
    )

    # Use `method_call` hook to pass flask config to each service method
    # context. But if you have any better ideas do it, make a pull request.
    # NOTE. I refuse idea to wrap each call into Flask application context
    # because in fact we inside Spyne app context, not the Flask one.
    def _flask_config_context(ctx):
        ctx.udc = UserDefinedContext(flask_app.config)
    application.event_manager.add_listener('method_call', _flask_config_context)

    return application



# encoding: utf8
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from pprint import pformat

from spyne.util import urlencode


def _start_response(code, headers):
    print(code, pformat(headers))

def call_wsgi_app_kwargs(app, _mn='some_call', _headers=None, **kwargs):
    return call_wsgi_app(app, _mn, _headers, kwargs.items())

def call_wsgi_app(app, mn='some_call', headers=None, body_pairs=None):
    if headers is None:
        headers = {}
    if body_pairs is None:
        body_pairs = []

    body_pairs = [(k,str(v)) for k,v in body_pairs]

    request = {
        'QUERY_STRING': urlencode(body_pairs),
        'PATH_INFO': '/%s' % mn,
        'REQUEST_METHOD': 'GET',
        'SERVER_NAME': 'spyne.test',
        'SERVER_PORT': '0',
        'wsgi.url_scheme': 'http',
    }

    print(headers)
    request.update(headers)

    out_string = []
    t = None
    for s in app(request, _start_response):
        t = type(s)
        out_string.append(s)

    if t == bytes:
        out_string = b''.join(out_string)
    else:
        out_string = ''.join(out_string)

    return out_string

from os import mkdir, getcwd
from os.path import join, basename


def show(elt, tn=None, stdout=True):
    if tn is None:
        import inspect

        for frame in inspect.stack():
            if frame[3].startswith("test_"):
                cn = frame[0].f_locals['self'].__class__.__name__
                tn = "%s.%s" % (cn, frame[3])
                break

        else:
            raise Exception("don't be lazy and pass test name.")

    from lxml import html, etree
    out_string = etree.tostring(elt, pretty_print=True)
    if stdout:
        print(out_string)

    fn = '%s.html' % tn
    if basename(getcwd()) != 'test_html':
        try:
            mkdir('test_html')
        except OSError:
            pass

        f = open(join("test_html", fn), 'wb')
    else:
        f = open(fn, 'wb')

    f.write(html.tostring(elt, pretty_print=True, doctype="<!DOCTYPE html>"))
